#!/usr/bin/env python
"""
udocker unit tests: FileUtil
"""

import sys
import os
from unittest import TestCase, main
from unittest.mock import patch, mock_open
from udocker.utils.fileutil import FileUtil
from udocker.config import Config

STDOUT = sys.stdout
STDERR = sys.stderr
UDOCKER_TOPDIR = "test_topdir"
BUILTINS = "builtins"


def find_str(self, find_exp, where):
    """Find string in test output messages."""
    found = False
    for item in where:
        if find_exp in str(item):
            self.assertTrue(True)
            found = True
            break
    if not found:
        self.assertTrue(False)


def is_writable_file(obj):
    """Check if obj is a file."""
    try:
        obj.write("")
    except(AttributeError, OSError, IOError):
        return False
    else:
        return True


class FileUtilTestCase(TestCase):
    """Test FileUtil() file manipulation methods."""

    def setUp(self):
        Config().getconf()
        Config().conf["tmpdir"] = "/tmp"

    def tearDown(self):
        pass

    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_01_init(self, mock_regpre, mock_base, mock_absp):
        """Test01 FileUtil() constructor."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        futil = FileUtil('filename.txt')
        self.assertEqual(futil.filename, os.path.abspath('filename.txt'))
        self.assertTrue(mock_regpre.called)

        futil = FileUtil('-')
        self.assertEqual(futil.filename, '-')
        self.assertEqual(futil.basename, '-')

    # @patch('udocker.utils.fileutil.os.path.isdir')
    # @patch('udocker.utils.fileutil.os.path.realpath')
    # def test_02__register_prefix(self, mock_rpath,
    #                              mock_isdir):
    #     """Test02 FileUtil._register_prefix()."""
    #     prefix = "/dir"
    #     res = ["/dir/", "/dir/", "/dir/", "/dir/"]
    #     mock_rpath.return_value = "/dir"
    #     mock_isdir.return_value = True
    #     futil = FileUtil('filename.txt')
    #     futil._register_prefix(prefix)
    #     self.assertEqual(futil.safe_prefixes, res)

    #     prefix = "/dir/"
    #     res = ["/dir/", "/dir/", "/dir/", "/dir/"]
    #     mock_rpath.return_value = "/dir/"
    #     mock_isdir.return_value = False
    #     futil = FileUtil('filename.txt')
    #     futil._register_prefix(prefix)
    #     self.assertEqual(futil.safe_prefixes, res)

    @patch.object(FileUtil, '_register_prefix')
    def test_03_register_prefix(self, mock_regpre):
        """Test03 FileUtil.register_prefix()."""
        mock_regpre.return_value = None
        futil = FileUtil('filename.txt')
        futil.register_prefix()
        self.assertTrue(mock_regpre.called)


    @patch('udocker.utils.fileutil.os.umask')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_04_umask(self, mock_regpre, mock_base,
                      mock_absp, mock_umask):
        """Test04 FileUtil.umask()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_umask.return_value = 0
        futil = FileUtil("somedir")
        status = futil.umask()
        self.assertTrue(status)

        mock_umask.return_value = 0
        futil = FileUtil("somedir")
        FileUtil.orig_umask = 0
        status = futil.umask(1)
        self.assertTrue(status)
        self.assertEqual(FileUtil.orig_umask, 0)

        mock_umask.return_value = 0
        futil = FileUtil("somedir")
        FileUtil.orig_umask = None
        status = futil.umask(1)
        self.assertTrue(status)
        self.assertEqual(FileUtil.orig_umask, 0)

    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_05_mktmp(self, mock_regpre, mock_base, mock_absp):
        """Test05 FileUtil.mktmp()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename2.txt'
        mock_absp.return_value = '/tmp/filename2.txt'
        Config().conf['tmpdir'] = '/somewhere'
        tmp_file = FileUtil('filename2.txt').mktmp()
        self.assertTrue(tmp_file.endswith('-filename2.txt'))
        self.assertTrue(tmp_file.startswith('/somewhere/udocker-'))
        self.assertGreater(len(tmp_file.strip()), 68)

    @patch('udocker.utils.fileutil.os.makedirs')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_06_mkdir(self, mock_regpre, mock_base,
                      mock_absp, mock_mkdirs):
        """Test06 FileUtil.mkdir()"""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_mkdirs.return_value = True
        status = FileUtil("somedir").mkdir()
        self.assertTrue(status)

        mock_mkdirs.side_effect = OSError("fail")
        status = FileUtil("somedir").mkdir()
        self.assertFalse(status)

    @patch('udocker.utils.fileutil.os.rmdir')
    def test_07_rmdir(self, mock_rmdir):
        """Test07 FileUtil.rmdir()."""
        mock_rmdir.return_value = None
        status = FileUtil("somedir").rmdir()
        self.assertTrue(status)

        mock_rmdir.side_effect = OSError("fail")
        status = FileUtil("somedir").rmdir()
        self.assertFalse(status)

    @patch.object(FileUtil, 'mktmp')
    @patch.object(FileUtil, 'mkdir')
    def test_08_mktmpdir(self, mock_mkdir, mock_mktmp):
        """Test08 FileUtil.mktmpdir()."""
        mock_mktmp.return_value = "/dir"
        mock_mkdir.return_value = True
        status = FileUtil("somedir").mktmpdir()
        self.assertEqual(status, "/dir")

        mock_mktmp.return_value = "/dir"
        mock_mkdir.return_value = False
        status = FileUtil("somedir").mktmpdir()
        self.assertEqual(status, None)

    @patch('udocker.utils.fileutil.os.lstat')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_09_uid(self, mock_regpre, mock_base,
                    mock_absp, mock_stat):
        """Test09 FileUtil.uid()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_stat.return_value.st_uid = 1234
        Config().conf['tmpdir'] = "/tmp"
        futil = FileUtil("filename.txt")
        uid = futil.uid()
        self.assertEqual(uid, 1234)

        mock_stat.side_effect = OSError("fail")
        Config().conf['tmpdir'] = "/tmp"
        futil = FileUtil("filename.txt")
        uid = futil.uid()
        self.assertEqual(uid, -1)

    @patch('udocker.utils.fileutil.os.path.isdir')
    @patch('udocker.utils.fileutil.os.path.realpath')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_10__is_safe_prefix(self, mock_regpre, mock_base,
                                mock_absp, mock_rpath, mock_isdir):
        """Test10 FileUtil._is_safe_prefix()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_rpath.return_value = '/tmp/filename.txt'
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = []
        status = futil._is_safe_prefix("/AAA")
        self.assertFalse(status)

        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_rpath.return_value = '/tmp/filename.txt'
        mock_isdir.return_value = True
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = ["/tmp"]
        status = futil._is_safe_prefix("/tmp/AAA")
        self.assertTrue(status)

    # TODO: The is a bug, calls _chmod, and _chown not implemented
    # @patch.object(FileUtil, '_chmod')
    # @patch('udocker.utils.fileutil.os.walk')
    # @patch('udocker.utils.fileutil.os.lchown')
    # @patch('udocker.utils.fileutil.os.path.abspath')
    # @patch('udocker.utils.fileutil.os.path.basename')
    # @patch.object(FileUtil, '_register_prefix')
    # def test_11_chown(self, mock_regpre, mock_base,
    #                   mock_absp, mock_lchown, mock_walk,
    #                   mock_fuchmod):
    #     """Test11 FileUtil.chown()."""
    #     mock_regpre.return_value = None
    #     mock_base.return_value = 'filename.txt'
    #     mock_absp.return_value = '/tmp/filename.txt'
    #     mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
    #     mock_lchown.side_effect = [None, None, None, None]
    #     mock_fuchmod.return_value = None
    #     futil = FileUtil("somedir")
    #     FileUtil.safe_prefixes = ["/tmp"]
    #     status = futil.chown(0, 0, False)
    #     self.assertTrue(status)
    #     self.assertTrue(mock_fuchmod.called)
    #     self.assertFalse(mock_walk.called)

    #     mock_regpre.return_value = None
    #     mock_base.return_value = 'filename.txt'
    #     mock_absp.return_value = '/tmp/filename.txt'
    #     mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
    #     mock_lchown.side_effect = [None, None, None, None]
    #     mock_fuchmod.return_value = None
    #     futil = FileUtil("somedir")
    #     FileUtil.safe_prefixes = ["/tmp"]
    #     status = futil.chown(0, 0, True)
    #     self.assertTrue(status)
    #     self.assertTrue(mock_fuchmod.called)
    #     self.assertTrue(mock_walk.called)
    #     self.assertTrue(mock_lchown.called)

    #     mock_regpre.return_value = None
    #     mock_base.return_value = 'filename.txt'
    #     mock_absp.return_value = '/tmp/filename.txt'
    #     mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
    #     mock_lchown.side_effect = [None, None, None, None]
    #     mock_fuchmod.side_effect = OSError("fail")
    #     futil = FileUtil("somedir")
    #     FileUtil.safe_prefixes = ["/tmp"]
    #     status = futil.chown(0, 0, False)
    #     self.assertFalse(status)

    @patch.object(FileUtil, 'chown')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_12_rchown(self, mock_regpre, mock_base,
                       mock_absp, mock_fuchown):
        """Test12 FileUtil.rchown()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_fuchown.return_value = True
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = ["/tmp"]
        status = futil.rchown()
        self.assertTrue(status)

    @patch('udocker.utils.fileutil.stat')
    @patch('udocker.utils.fileutil.os.chmod')
    @patch('udocker.utils.fileutil.os.lstat')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_13__chmod(self, mock_regpre, mock_base,
                       mock_absp, mock_lstat, mock_chmod,
                       mock_stat):
        """Test13 FileUtil._chmod()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_lstat.return_value.st_mode = 33277
        mock_chmod.return_value = None
        mock_stat.return_value.S_ISREG = True
        mock_stat.return_value.S_IMODE = 509
        futil = FileUtil("somedir")
        futil._chmod("somefile")
        self.assertTrue(mock_lstat.called)
        self.assertTrue(mock_stat.S_ISREG.called)
        self.assertTrue(mock_stat.S_IMODE.called)

        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_lstat.return_value.st_mode = 33277
        mock_chmod.return_value = None
        mock_stat.return_value.S_ISREG = False
        mock_stat.return_value.S_ISDIR = True
        mock_stat.return_value.S_IMODE = 509
        futil = FileUtil("somedir")
        futil._chmod("somefile")
        self.assertTrue(mock_chmod.called)
        self.assertTrue(mock_stat.S_IMODE.called)

    @patch.object(FileUtil, '_chmod')
    @patch('udocker.utils.fileutil.os.walk')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test14_chmod(self, mock_regpre, mock_base,
                     mock_absp, mock_walk, mock_fuchmod):
        """Test14 FileUtil.chmod()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
        mock_fuchmod.return_value = None
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = ["/tmp"]
        status = futil.chmod(0o600, 0o700, 0o755, False)
        self.assertTrue(status)
        self.assertTrue(mock_fuchmod.called)
        self.assertFalse(mock_walk.called)

        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
        mock_fuchmod.side_effect = [None, None, None, None]
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = ["/tmp"]
        status = futil.chmod(0o600, 0o700, 0o755, True)
        self.assertTrue(status)
        self.assertTrue(mock_fuchmod.called)
        self.assertTrue(mock_walk.called)

        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
        mock_fuchmod.side_effect = OSError("fail")
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = ["/tmp"]
        status = futil.chmod(0o600, 0o700, 0o755, False)
        self.assertFalse(status)

    @patch.object(FileUtil, 'chmod')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_15_rchmod(self, mock_regpre, mock_base,
                       mock_absp, mock_fuchmod):
        """Test15 FileUtil.rchmod()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_fuchmod.return_value = True
        futil = FileUtil("somedir")
        FileUtil.safe_prefixes = ["/tmp"]
        futil.rchmod()
        self.assertTrue(mock_fuchmod.called)

    @patch('udocker.utils.fileutil.os.rmdir')
    @patch('udocker.utils.fileutil.os.unlink')
    @patch('udocker.utils.fileutil.os.chmod')
    @patch('udocker.utils.fileutil.os.walk')
    @patch('udocker.utils.fileutil.os.path.islink')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_16__removedir(self, mock_regpre, mock_base, mock_absp,
                           mock_islink, mock_walk, mock_chmod,
                           mock_unlink, mock_rmdir):
        """Test16 FileUtil._removedir()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_walk.return_value = [("/tmp", ["dir"], ["file"]), ]
        mock_islink.side_effect = [False, True, True, True]
        mock_chmod.side_effect = [None, None, None, None]
        mock_unlink.side_effect = [None, None, None, None]
        mock_rmdir.side_effect = [None, None, None, None]
        # remove directory under /tmp OK
        futil = FileUtil("/tmp/directory")
        status = futil._removedir()
        self.assertTrue(mock_walk.called)
        self.assertTrue(mock_islink.call_count, 2)
        self.assertTrue(mock_chmod.call_count, 3)
        self.assertTrue(status)

        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_walk.return_value = list()
        mock_chmod.side_effect = OSError("fail")
        futil = FileUtil("/directory")
        status = futil._removedir()
        self.assertFalse(status)

    @patch('udocker.utils.fileutil.os.path.realpath')
    @patch('udocker.utils.fileutil.os.path.exists')
    @patch('udocker.utils.fileutil.Msg')
    @patch('udocker.utils.fileutil.os.remove')
    @patch('udocker.utils.fileutil.os.path.islink')
    @patch('udocker.utils.fileutil.os.path.isfile')
    @patch('udocker.utils.fileutil.os.path.isdir')
    @patch.object(FileUtil, 'uid')
    @patch.object(FileUtil, '_is_safe_prefix')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_17_remove(self, mock_regpre, mock_base, mock_absp, mock_safe,
                       mock_uid, mock_isdir,
                       mock_isfile, mock_islink, mock_remove, mock_msg,
                       mock_exists, mock_realpath):
        """Test17 FileUtil.remove()."""
        mock_msg.level = 0
        mock_regpre.return_value = None
        mock_base.return_value = '/filename4.txt'
        mock_absp.return_value = '/filename4.txt'
        mock_uid.return_value = 1000
        # file does not exist (regression of #50)
        mock_isdir.return_value = True
        mock_isfile.return_value = True
        mock_exists.return_value = True
        mock_safe.return_value = True
        mock_islink.return_value = False
        mock_remove.return_value = None
        Config().conf['uid'] = 1000
        Config().conf['tmpdir'] = "/tmp"
        mock_realpath.return_value = "/tmp"
        # under /
        futil = FileUtil("/filename4.txt")
        status = futil.remove()
        self.assertTrue(status)

        # wrong uid
        mock_base.return_value = 'filename4.txt'
        mock_absp.return_value = '/tmp/filename4.txt'
        mock_uid.return_value = 1001
        futil = FileUtil("/tmp/filename4.txt")
        status = futil.remove()
        self.assertTrue(status)

        # under /tmp TEST to be checked
        # mock_base.return_value = 'filename4.txt'
        # mock_absp.return_value = '/tmp/filename4.txt'
        # mock_uid.return_value = 1000
        # futil = FileUtil("/tmp/filename4.txt")
        # status = futil.remove()
        # self.assertTrue(status)

        # under user home TEST to be checked
        # mock_base.return_value = 'filename4.txt'
        # mock_absp.return_value = '/home/user/.udocker/filename4.txt'
        # futil = FileUtil("/home/user/.udocker/filename4.txt")
        # futil.safe_prefixes.append("/home/user/.udocker")
        # status = futil.remove()
        # self.assertTrue(status)

        # outside of scope 1
        mock_base.return_value = 'filename4.txt'
        mock_absp.return_value = '/etc/filename4.txt'
        mock_safe.return_value = False
        futil = FileUtil("/etc/filename4.txt")
        futil.safe_prefixes = []
        status = futil.remove()
        self.assertTrue(status)

    @patch('udocker.utils.fileutil.Msg')
    @patch('udocker.utils.fileutil.Uprocess.call')
    @patch('udocker.utils.fileutil.os.path.isfile')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_18_verify_tar(self, mock_regpre, mock_base, mock_absp,
                           mock_isfile, mock_call, mock_msg):
        """Test18 FileUtil.verify_tar()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_msg.level = 0
        mock_msg.VER = 4
        mock_isfile.return_value = False
        mock_call.return_value = 0
        status = FileUtil("tarball.tar").verify_tar()
        self.assertFalse(status)

        mock_isfile.return_value = True
        mock_call.return_value = 0
        status = FileUtil("tarball.tar").verify_tar()
        self.assertTrue(status)

        mock_isfile.return_value = True
        mock_call.return_value = 1
        status = FileUtil("tarball.tar").verify_tar()
        self.assertFalse(status)

    @patch('udocker.utils.fileutil.Msg')
    @patch('udocker.utils.fileutil.Uprocess.call')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_19_tar(self, mock_regpre, mock_base, mock_absp,
                    mock_call, mock_msg):
        """Test19 FileUtil.tar()."""
        mock_msg.level = 0
        mock_msg.VER = 4
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_call.return_value = 1
        status = FileUtil("tarball.tar").tar("tarball.tar")
        self.assertFalse(status)

        mock_call.return_value = 0
        status = FileUtil("tarball.tar").tar("tarball.tar")
        self.assertTrue(status)

    @patch('udocker.utils.fileutil.Uprocess.pipe')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_20_copydir(self, mock_regpre, mock_base, mock_absp,
                        mock_call):
        """Test20 FileUtil.copydir()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_call.return_value = 1
        status = FileUtil("filename.txt").copydir("/dir1")
        self.assertEqual(status, 1)

        mock_call.return_value = 0
        status = FileUtil("filename.txt").copydir("/dir1")
        self.assertEqual(status, 0)

    @patch.object(FileUtil, 'remove')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_21_cleanup(self, mock_regpre, mock_base, mock_absp, mock_remove):
        """Test21 FileUtil.cleanup()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        Config().conf['tmpdir'] = "/tmp"
        FileUtil.tmptrash = {'file1.txt': None, 'file2.txt': None}
        FileUtil("").cleanup()
        self.assertEqual(mock_remove.call_count, 2)

    @patch('udocker.utils.fileutil.os.path.isdir')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_22_isdir(self, mock_regpre, mock_base,
                      mock_absp, mock_isdir):
        """Test22 FileUtil.isdir()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_isdir.return_value = True
        status = FileUtil("filename.txt").isdir()
        self.assertTrue(status)

        mock_isdir.return_value = False
        status = FileUtil("filename.txt").isdir()
        self.assertFalse(status)

    @patch('udocker.utils.fileutil.os.stat')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_23_size(self, mock_regpre, mock_base, mock_absp, mock_stat):
        """Test23 FileUtil.size()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_stat.return_value.st_size = 4321
        size = FileUtil("somefile").size()
        self.assertEqual(size, 4321)

        mock_stat.side_effect = OSError("fail")
        size = FileUtil("somefile").size()
        self.assertEqual(size, -1)

    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_24_getdata(self, mock_regpre, mock_base, mock_absp):
        """Test24 FileUtil.getdata()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        with patch(BUILTINS + '.open', mock_open(read_data='qwerty')):
            data = FileUtil("somefile").getdata()
            self.assertEqual(data, 'qwerty')

    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_25_get1stline(self, mock_regpre, mock_base, mock_absp):
        """Test25 FileUtil.get1stline()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        with patch(BUILTINS + '.open', mock_open(read_data='qwerty')):
            data = FileUtil("somefile").get1stline()
            self.assertEqual(data, 'qwerty')

    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_26_putdata(self, mock_regpre, mock_base, mock_absp):
        """Test26 FileUtil.putdata()"""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        futil = FileUtil("somefile")
        futil.filename = ""
        data = futil.putdata("qwerty")
        self.assertFalse(data)

        with patch(BUILTINS + '.open', mock_open()):
            data = FileUtil("somefile").putdata("qwerty")
            self.assertEqual(data, 'qwerty')

    @patch('udocker.utils.fileutil.os.path.split')
    @patch('udocker.utils.fileutil.os.path.exists')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_27_getvalid_path(self, mock_regpre, mock_base,
                              mock_absp, mock_exists, mock_split):
        """Test27 FileUtil.getvalid_path()."""
        mock_regpre.return_value = None
        mock_base.return_value = "somefile"
        mock_absp.return_value = "somefile"
        mock_exists.return_value = True
        futil = FileUtil("somefile")
        status = futil.getvalid_path()
        self.assertEqual(status, "somefile")

        mock_regpre.return_value = None
        mock_base.return_value = "somefile"
        mock_absp.return_value = "somefile"
        mock_exists.side_effect = [False, True]
        mock_split.return_value = ("somefile", "/dir")
        futil = FileUtil("somefile")
        status = futil.getvalid_path()
        self.assertEqual(status, "somefile")

    @patch('udocker.utils.fileutil.os.path.islink')
    @patch('udocker.utils.fileutil.os.path.normpath')
    @patch('udocker.utils.fileutil.os.path.realpath')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_28__cont2host(self, mock_regpre, mock_base,
                           mock_absp, mock_rpath, mock_normp,
                           mock_islink):
        """Test28 FileUtil._cont2host()."""
        hpath = ""
        croot = "/ROOT"
        mock_regpre.return_value = None
        mock_base.return_value = "somefile"
        mock_absp.return_value = "/home/somefile"
        futil = FileUtil("somefile")
        status = futil._cont2host(hpath, croot)
        self.assertEqual(status, "")

        hpath = "/usr/bin"
        croot = "/ROOT/usr/bin"
        mock_regpre.return_value = None
        mock_base.return_value = "somefile"
        mock_absp.return_value = "/home/somefile"
        mock_rpath.return_value = "/ROOT/usr/bin"
        mock_normp.return_value = "/ROOT/usr/bin"
        mock_islink.return_value = False
        futil = FileUtil("somefile")
        status = futil._cont2host(hpath, croot)
        self.assertEqual(status, "/ROOT/usr/bin")

        hpath = "/usr/bin"
        croot = "/ROOT/usr/bin"
        vol = ["/home/user:/ROOT/home/user"]
        mock_regpre.return_value = None
        mock_base.return_value = "somefile"
        mock_absp.return_value = "/home/somefile"
        mock_rpath.return_value = "/ROOT/usr/bin"
        mock_normp.return_value = "/ROOT/usr/bin"
        mock_islink.return_value = False
        futil = FileUtil("somefile")
        status = futil._cont2host(hpath, croot, vol)
        self.assertEqual(status, "/ROOT/usr/bin")

    @patch.object(FileUtil, '_cont2host')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_29_cont2host(self, mock_regpre, mock_base,
                          mock_absp, mock_c2h):
        """Test29 FileUtil.cont2host()."""
        mock_regpre.return_value = None
        mock_base.return_value = "somefile"
        mock_absp.return_value = "somefile"
        mock_c2h.return_value = "/ROOT/dir"
        futil = FileUtil("somefile")
        status = futil.cont2host("/ROOT/dir")
        self.assertEqual(status, "/ROOT/dir")
        self.assertTrue(mock_c2h.called)

    @patch('udocker.utils.fileutil.os.access')
    @patch('udocker.utils.fileutil.os.path.isfile')
    @patch.object(FileUtil, '_cont2host')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_30__find_exec(self, mock_regpre, mock_base, mock_absp,
                           mock_c2h, mock_isfile, mock_access):
        """Test30 FileUtil._find_exec()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        futil = FileUtil("somefile")
        status = futil._find_exec("")
        self.assertEqual(status, "")

        mock_isfile.return_value = True
        mock_access.return_value = True
        futil = FileUtil("/bin/ls")
        status = futil._find_exec("/bin")
        self.assertEqual(status, "/bin/ls")

        mock_isfile.return_value = True
        mock_access.return_value = True
        mock_c2h.return_value = "/ROOT/bin/ls"
        futil = FileUtil("/bin/ls")
        status = futil._find_exec("/bin", "", "", ".", False)
        self.assertEqual(status, "/bin/ls")

    @patch.object(FileUtil, '_find_exec')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_31_find_exec(self, mock_regpre, mock_base, mock_absp,
                          mock_findexe):
        """Test31 FileUtil.find_exec()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'ls'
        mock_absp.return_value = '/bin/ls'
        mock_findexe.return_value = '/bin/ls'
        futil = FileUtil("/bin/ls")
        status = futil.find_exec("/bin", "", "", ".", False)
        self.assertEqual(status, "/bin/ls")

    @patch('udocker.utils.fileutil.os.rename')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_32_rename(self, mock_regpre, mock_base, mock_absp,
                       mock_rename):
        """Test32 FileUtil.rename()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'ls'
        mock_absp.return_value = '/bin/ls'
        mock_rename.return_value = None
        futil = FileUtil("/bin/ls")
        status = futil.rename("somefile")
        self.assertTrue(status)

        mock_regpre.return_value = None
        mock_base.return_value = 'ls'
        mock_absp.return_value = '/bin/ls'
        mock_rename.side_effect = OSError("fail")
        futil = FileUtil("/bin/ls")
        status = futil.rename("somefile")
        self.assertFalse(status)

    # def test_33__stream2file(self):
    #     """Test33 FileUtil._stream2file()."""

    # def test_34__file2stream(self):
    #     """Test34 FileUtil._file2stream()."""

    # def test_35__file2file(self):
    #     """Test35 FileUtil._file2file()."""

    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_36_copyto(self, mock_regpre, mock_base, mock_absp):
        """Test36 FileUtil.copyto()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        with patch(BUILTINS + '.open', mock_open()):
            status = FileUtil("source").copyto("dest")
            self.assertTrue(status)

            status = FileUtil("source").copyto("dest", "w")
            self.assertTrue(status)

            status = FileUtil("source").copyto("dest", "a")
            self.assertTrue(status)

    @patch('udocker.utils.fileutil.os.path.exists')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_37_find_file_in_dir(self, mock_regpre, mock_base, mock_absp,
                                 mock_exists):
        """Test37 FileUtil.find_file_in_dir()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/dir'
        file_list = []
        status = FileUtil("/dir").find_file_in_dir(file_list)
        self.assertEqual(status, "")

        file_list = ["F1", "F2"]
        mock_exists.side_effect = [False, False]
        status = FileUtil("/dir").find_file_in_dir(file_list)
        self.assertEqual(status, "")

        file_list = ["F1", "F2"]
        mock_exists.side_effect = [False, True]
        status = FileUtil("/dir").find_file_in_dir(file_list)
        self.assertEqual(status, "/dir/F2")

    @patch('udocker.utils.fileutil.os.symlink')
    @patch('udocker.utils.fileutil.os.remove')
    @patch('udocker.utils.fileutil.os.chmod')
    @patch('udocker.utils.fileutil.os.access')
    @patch('udocker.utils.fileutil.os.path.realpath')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_38__link_change_apply(self, mock_regpre, mock_base, mock_absp,
                                   mock_realpath, mock_access, mock_chmod,
                                   mock_remove, mock_symlink):
        """Test38 FileUtil._link_change_apply()."""
        mock_regpre.return_value = None
        mock_chmod.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_realpath.return_value = "/HOST/DIR"
        mock_access.return_value = True
        futil = FileUtil("/con")
        futil._link_change_apply("/con/lnk_new", "/con/lnk", False)
        self.assertTrue(mock_remove.called)
        self.assertTrue(mock_symlink.called)

        # mock_regpre.return_value = None
        # mock_base.return_value = 'filename.txt'
        # mock_absp.return_value = '/tmp/filename.txt'
        # mock_realpath.return_value = "/HOST/DIR"
        # mock_access.return_value = False
        # mock_chmod.side_effect = [None, None]
        # mock_remove.return_value = None
        # mock_symlink.return_value = None
        # mock_realpath.return_value = "/HOST/DIR"
        # futil = FileUtil("/con")
        # futil._link_change_apply("/con/lnk_new", "/con/lnk", True)
        # self.assertTrue(mock_chmod.call_count, 2)
        # self.assertTrue(mock_remove.called)
        # self.assertTrue(mock_symlink.called)

    @patch('udocker.utils.fileutil.os.access')
    @patch('udocker.utils.fileutil.os.readlink')
    @patch.object(FileUtil, '_link_change_apply', return_value=None)
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_39__link_set(self, mock_regpre, mock_base, mock_absp,
                          mock_lnchange, mock_readlink, mock_access):
        """Test39 FileUtil._link_set()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_readlink.return_value = "X"
        futil = FileUtil("/con")
        status = futil._link_set("/con/lnk", "", "/con", False)
        self.assertFalse(status)

        mock_readlink.return_value = "/con"
        futil = FileUtil("/con")
        status = futil._link_set("/con/lnk", "", "/con", False)
        self.assertFalse(status)

        mock_readlink.return_value = "/HOST/DIR"
        futil = FileUtil("/con")
        status = futil._link_set("/con/lnk", "", "/con", False)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

        mock_readlink.return_value = "/HOST/DIR"
        mock_access.return_value = True
        futil = FileUtil("/con")
        status = futil._link_set("/con/lnk", "", "/con", True)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

        mock_readlink.return_value = "/HOST/DIR"
        mock_access.return_value = False
        futil = FileUtil("/con")
        status = futil._link_set("/con/lnk", "", "/con", True)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

    @patch('udocker.utils.fileutil.os.access')
    @patch('udocker.utils.fileutil.os.readlink')
    @patch.object(FileUtil, '_link_change_apply', return_value=None)
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_40__link_restore(self, mock_regpre, mock_base, mock_absp,
                              mock_lnchange, mock_readlink, mock_access):
        """Test40 FileUtil._link_restore()."""
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_readlink.return_value = "/con/AAA"
        futil = FileUtil("/con")
        status = futil._link_restore("/con/lnk", "/con", "/root", False)
        self.assertTrue(status)

        mock_readlink.return_value = "/con/AAA"
        futil = FileUtil("/con")
        status = futil._link_restore("/con/lnk", "/con", "/root", False)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

        mock_readlink.return_value = "/root/BBB"
        futil = FileUtil("/con")
        status = futil._link_restore("/con/lnk", "/con", "/root", False)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

        mock_readlink.return_value = "/XXX"
        futil = FileUtil("/con")
        status = futil._link_restore("/con/lnk", "/con", "/root", False)
        self.assertTrue(mock_lnchange.called)
        self.assertFalse(status)

        mock_readlink.return_value = "/root/BBB"
        futil = FileUtil("/con")
        status = futil._link_restore("/con/lnk", "/con", "/root", True)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

        mock_readlink.return_value = "/root/BBB"
        mock_access.return_value = False
        futil = FileUtil("/con")
        status = futil._link_restore("/con/lnk", "", "/root", True)
        self.assertTrue(mock_lnchange.called)
        self.assertTrue(status)

    @patch.object(FileUtil, '_link_restore')
    @patch.object(FileUtil, '_link_set')
    @patch('udocker.utils.fileutil.Msg')
    @patch.object(FileUtil, '_is_safe_prefix')
    @patch('udocker.utils.fileutil.os.lstat')
    @patch('udocker.utils.fileutil.os.path.islink')
    @patch('udocker.utils.fileutil.os.walk')
    @patch('udocker.utils.fileutil.os.path.realpath')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_41_links_conv(self, mock_regpre, mock_base, mock_absp,
                           mock_realpath, mock_walk, mock_islink,
                           mock_lstat, mock_is_safe_prefix, mock_msg,
                           mock_link_set, mock_link_restore):
        """Test41 FileUtil.links_conv()."""
        mock_msg.level = 0
        mock_regpre.return_value = None
        mock_base.return_value = 'filename.txt'
        mock_absp.return_value = '/tmp/filename.txt'
        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = False
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, True, "")
        self.assertEqual(status, None)

        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = True
        mock_walk.return_value = []
        mock_islink.return_value = True
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, True, "")
        self.assertEqual(status, [])

        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = True
        mock_walk.return_value = [("/", [], []), ]
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, True, "")
        self.assertEqual(status, [])

        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = True
        mock_islink = False
        mock_walk.return_value = [("/", [], ["F1", "F2"]), ]
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, True, "")
        self.assertEqual(status, [])

        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = True
        mock_islink = True
        mock_lstat.return_value.st_uid = 1
        Config().conf['uid'] = 0
        mock_walk.return_value = [("/", [], ["F1", "F2"]), ]
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, True, "")
        self.assertEqual(status, [])

        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = True
        mock_islink = True
        mock_lstat.return_value.st_uid = 1
        mock_link_set.reset_mock()
        mock_link_restore.reset_mock()
        Config().conf['uid'] = 1
        mock_walk.return_value = [("/", [], ["F1", "F2"]), ]
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, True, "")
        self.assertFalse(mock_link_restore.called)

        mock_realpath.return_value = "/ROOT"
        mock_is_safe_prefix.return_value = True
        mock_islink = True
        mock_lstat.return_value.st_uid = 1
        mock_link_set.reset_mock()
        mock_link_restore.reset_mock()
        Config().conf['uid'] = 1
        mock_walk.return_value = [("/", [], ["F1", "F2"]), ]
        futil = FileUtil("/ROOT")
        status = futil.links_conv(False, False, "")
        self.assertFalse(mock_link_set.called)

    @patch('udocker.utils.fileutil.os.path.isdir')
    @patch('udocker.utils.fileutil.os.listdir')
    @patch('udocker.utils.fileutil.os.path.dirname')
    @patch('udocker.utils.fileutil.os.path.abspath')
    @patch('udocker.utils.fileutil.os.path.basename')
    @patch.object(FileUtil, '_register_prefix')
    def test_42_match(self, mock_regpre, mock_base, mock_absp,
                      mock_dirname, mock_listdir, mock_isdir):
        """Test42 FileUtil.match()."""
        mock_regpre.return_value = None
        mock_base.return_value = "/con/filename.txt"
        mock_absp.return_value = "/con/filename.txt"
        mock_dirname.return_value = "/con/fil*"
        mock_isdir.return_value = False
        mock_listdir.return_value = list()
        futil = FileUtil("/con/filename.txt")
        status = futil.match()
        self.assertEqual(status, [])

        # mock_regpre.return_value = None
        # mock_base.return_value = "fil*"
        # mock_absp.return_value = "/con/filename*"
        # mock_dirname.return_value = "/con/"
        # mock_isdir.return_value = True
        # mock_listdir = ["filename1", "filename2"]
        # futil = FileUtil("/con/filename*")
        # status = futil.match()
        # self.assertEqual(status, ["/con/filename1", "/con/filename2"])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: Uenv
"""

from unittest import TestCase, main
from udocker.utils.uenv import Uenv


class UenvTestCase(TestCase):
    """Test Uenv()."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_01_init(self):
        """Test01 Uenv() constructor"""
        envt = 'HOME=/home/user'
        result = {'HOME': '/home/user'}
        self.assertEqual(Uenv(envt).env, result)

    # def test_02__iter__(self):
    #     """Test02 Uenv().__iter__"""

    def test_03__len__(self):
        """Test03 Uenv().__len__"""
        envt = 'HOME=/home/user'
        uenv = Uenv(envt)
        uenv.env = {'HOME': '/home/user', 'LANG': 'en_US.UTF-8'}
        self.assertEqual(uenv.__len__(), 2)

    def test_04_append(self):
        """Test04 Uenv().append"""
        envt = 'HOME=/home/user'
        env_app = 'LANG=en_US.UTF-8'
        result = {'HOME': '/home/user', 'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.append(env_app)
        self.assertEqual(uenv.env, result)

    def test_05_appendif(self):
        """Test05 Uenv().appendif"""
        envt = 'HOME=/home/user'
        env_app = 'LANG=en_US.UTF-8'
        result = {'HOME': '/home/user', 'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.appendif(env_app)
        self.assertEqual(uenv.env, result)

        envt = 'HOME=/home/user'
        env_app = envt
        result = {'HOME': '/home/user'}
        uenv = Uenv(envt)
        uenv.appendif(env_app)
        self.assertEqual(uenv.env, result)

    def test_06_extend(self):
        """Test06 Uenv().extend"""
        envt = 'HOME=/home/user'
        envlist = ['LANG=en_US.UTF-8', 'USER=myuser']
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8',
                  'USER': 'myuser'}
        uenv = Uenv(envt)
        uenv.extend(envlist)
        self.assertEqual(uenv.env, result)

        envdict = {'LANG': 'en_US.UTF-8', 'USER': 'myuser'}
        uenv = Uenv(envt)
        uenv.extend(envdict)
        self.assertEqual(uenv.env, result)

    def test_07_extendif(self):
        """Test07 Uenv().extendif"""
        envt = 'HOME=/home/user'
        envlist = ['LANG=en_US.UTF-8', 'USER=myuser']
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8',
                  'USER': 'myuser'}
        uenv = Uenv(envt)
        uenv.extendif(envlist)
        self.assertEqual(uenv.env, result)

        envdict = {'LANG': 'en_US.UTF-8', 'USER': 'myuser'}
        uenv = Uenv(envt)
        uenv.extendif(envdict)
        self.assertEqual(uenv.env, result)

        envt = 'HOME=/home/user'
        envlist = ['LANG=en_US.UTF-8', 'HOME=/home/user']
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.extendif(envlist)
        self.assertEqual(uenv.env, result)

    def test_08_add(self):
        """Test08 Uenv().add"""
        envt = 'HOME=/home/user'
        envlist = ['LANG=en_US.UTF-8']
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.add(envlist)
        self.assertEqual(uenv.env, result)

        envt = 'HOME=/home/user'
        envstr = 'LANG=en_US.UTF-8'
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.add(envstr)
        self.assertEqual(uenv.env, result)

    def test_09_getenv(self):
        """Test09 Uenv().getenv"""
        envt = 'HOME=/home/user'
        uenv = Uenv(envt)
        uenv.getenv('HOME')
        self.assertEqual(uenv.env, {'HOME': '/home/user'})

    def test_10_setenv(self):
        """Test10 Uenv().setenv"""
        envt = 'HOME=/home/user'
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.setenv('LANG', 'en_US.UTF-8')
        self.assertEqual(uenv.env, result)

    def test_11_unsetenv(self):
        """Test11 Uenv().unsetenv"""
        envt = 'HOME=/home/user'
        result = {'HOME': '/home/user'}
        uenv = Uenv(envt)
        uenv.setenv('LANG', 'en_US.UTF-8')

        uenv.unsetenv('LANG')
        self.assertEqual(uenv.env, result)

    def test_12_list(self):
        """Test12 Uenv().list"""
        envt = 'HOME=/home/user'
        result = ['LANG=en_US.UTF-8', envt]
        uenv = Uenv(envt)
        uenv.setenv('LANG', 'en_US.UTF-8')
        self.assertEqual(uenv.list().sort(), result.sort())

    def test_13_dict(self):
        """Test13 Uenv().dict"""
        envt = 'HOME=/home/user'
        result = {'HOME': '/home/user',
                  'LANG': 'en_US.UTF-8'}
        uenv = Uenv(envt)
        uenv.setenv('LANG', 'en_US.UTF-8')
        self.assertEqual(uenv.dict(), result)

    def test_14_keys(self):
        """Test14 Uenv().keys"""
        envt = 'HOME=/home/user'
        result = {'HOME': None, 'LANG': None}
        uenv = Uenv(envt)
        uenv.setenv('LANG', 'en_US.UTF-8')
        self.assertEqual(uenv.keys(), result.keys())


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: FakechrootEngine
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock
from udocker.config import Config
from udocker.engine.fakechroot import FakechrootEngine


class FakechrootEngineTestCase(TestCase):
    """Docker container execution engine using Fakechroot
    Provides a chroot like environment to run containers.
    Uses Fakechroot as chroot alternative.
    Inherits from ContainerEngine class
    """

    def setUp(self):
        Config().getconf()
        Config().conf['hostauth_list'] = ("/etc/passwd", "/etc/group")
        Config().conf['cmd'] = "/bin/bash"
        Config().conf['cpu_affinity_exec_tools'] = \
            (["numactl", "-C", "%s", "--", ], ["taskset", "-c", "%s", ])
        Config().conf['valid_host_env'] = "HOME"
        Config().conf['username'] = "user"
        Config().conf['userhome'] = "/"
        Config().conf['oskernel'] = "4.8.13"
        Config().conf['location'] = ""

        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

        str_exmode = 'udocker.engine.execmode.ExecutionMode'
        self.execmode = patch(str_exmode)
        self.xmode = self.execmode.start()
        self.mock_execmode = Mock()
        self.xmode.return_value = self.mock_execmode

    def tearDown(self):
        self.lrepo.stop()
        self.execmode.stop()

    def test_01__init(self):
        """Test01 FakechrootEngine Constructor."""
        ufake = FakechrootEngine(self.local, self.xmode)
        self.assertEqual(ufake._fakechroot_so, "")
        self.assertIsNone(ufake._elfpatcher)

    @patch('udocker.engine.fakechroot.Msg')
    @patch('udocker.engine.fakechroot.sys.exit')
    @patch('udocker.engine.fakechroot.OSInfo')
    @patch('udocker.engine.fakechroot.os.path.realpath')
    @patch('udocker.engine.fakechroot.os.path.exists')
    @patch('udocker.engine.fakechroot.FileUtil')
    def test_02_select_fakechroot_so(self, mock_futil, mock_exists,
                                     mock_rpath, mock_osinfo, mock_sysex, mock_msg):
        """Test02 FakechrootEngine.select_fakechroot_so."""
        Config().conf['fakechroot_so'] = "/s/fake1"
        mock_msg.level = 0
        mock_exists.return_value = True
        mock_rpath.return_value = "/s/fake1"
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake.select_fakechroot_so()
        self.assertEqual(out, "/s/fake1")

        Config().conf['fakechroot_so'] = ""
        mock_exists.return_value = True
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake.select_fakechroot_so()
        self.assertEqual(out, "/libfakechroot.so")

        Config().conf['fakechroot_so'] = ""
        mock_exists.return_value = False
        mock_osinfo.return_value.arch.return_value = "amd64"
        mock_osinfo.return_value.osdistribution.return_value = \
            ("linux", "4.8.1")
        mock_futil.return_value.find_file_in_dir.return_value = ""
        mock_sysex.return_value = 1
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake.select_fakechroot_so()
        self.assertTrue(mock_sysex.called)
        self.assertTrue(mock_futil.called)

        Config().conf['fakechroot_so'] = ""
        mock_exists.return_value = False
        mock_osinfo.return_value.arch.return_value = "amd64"
        mock_osinfo.return_value.osdistribution.return_value = \
            ("linux", "4.8.1")
        mock_futil.return_value.find_file_in_dir.return_value = \
            "/libfakechroot.so"
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake.select_fakechroot_so()
        self.assertTrue(mock_futil.called)
        self.assertEqual(out, "/libfakechroot.so")

        Config().conf['fakechroot_so'] = ""
        mock_exists.return_value = False
        mock_osinfo.return_value.arch.return_value = "i386"
        mock_osinfo.return_value.osdistribution.return_value = \
            ("linux", "4.8.1")
        mock_futil.return_value.find_file_in_dir.return_value = \
            "/libfakechroot.so"
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake.select_fakechroot_so()
        self.assertTrue(mock_futil.called)
        self.assertEqual(out, "/libfakechroot.so")

        Config().conf['fakechroot_so'] = ""
        mock_exists.return_value = False
        mock_osinfo.return_value.arch.return_value = "arm64"
        mock_osinfo.return_value.osdistribution.return_value = \
            ("linux", "4.8.1")
        mock_futil.return_value.find_file_in_dir.return_value = \
            "/libfakechroot.so"
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake.select_fakechroot_so()
        self.assertTrue(mock_futil.called)
        self.assertEqual(out, "/libfakechroot.so")

    @patch('udocker.engine.fakechroot.ExecutionEngineCommon._setup_container_user_noroot')
    def test_03__setup_container_user(self, mock_cmm):
        """Test03 FakechrootEngine._setup_container_user()."""
        mock_cmm.return_value = True
        ufake = FakechrootEngine(self.local, self.xmode)
        status = ufake._setup_container_user("someuser")
        self.assertTrue(mock_cmm.called)
        self.assertTrue(status)

    @patch('udocker.engine.fakechroot.Msg')
    def test_04__uid_check(self, mock_msg):
        """Test04 FakechrootEngine._uid_check()."""
        mock_msg.level = 3
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake.opt["user"] = "0"
        ufake._uid_check()
        self.assertTrue(mock_msg.return_value.out.called)

    def test_05__get_volume_bindings(self):
        """Test05 FakechrootEngine._get_volume_bindings()."""
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake.opt["vol"] = ()
        status = ufake._get_volume_bindings()
        self.assertEqual(status, ('', ''))

        # ufake = FakechrootEngine(self.local, self.xmode)
        # ufake.opt["vol"] = ("/tmp", "/bin",)
        # status = ufake._get_volume_bindings()
        # self.assertEqual(status, ('', '/tmp!/tmp:/bin!/bin'))

    @patch('udocker.engine.fakechroot.os.path.exists')
    @patch('udocker.engine.fakechroot.FileUtil.cont2host')
    def test_06__get_access_filesok(self, mock_cont2host, mock_exists):
        """Test06 FakechrootEngine._get_access_filesok()."""
        Config.conf['access_files'] = ("/sys/infin",)
        mock_cont2host.return_value = "/c/sys/infin"
        mock_exists.return_value = False
        ufake = FakechrootEngine(self.local, self.xmode)
        out = ufake._get_access_filesok()
        self.assertEqual(out, "")

        Config.conf['access_files'] = ("/sys/infin",)
        mock_cont2host.return_value = "/c/sys/infin"
        mock_exists.return_value = True
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake.opt["vol"] = ("/tmp",)
        out = ufake._get_access_filesok()
        self.assertTrue(mock_cont2host.called)
        self.assertTrue(mock_exists.called)
        self.assertEqual(out, "/sys/infin")

    @patch.object(FakechrootEngine, '_get_volume_bindings')
    @patch.object(FakechrootEngine, 'select_fakechroot_so')
    @patch.object(FakechrootEngine, '_get_access_filesok')
    @patch.object(FakechrootEngine, '_is_volume')
    @patch('udocker.engine.fakechroot.os.path.realpath')
    @patch('udocker.engine.fakechroot.ElfPatcher')
    @patch('udocker.engine.fakechroot.FileUtil')
    def test_07__fakechroot_env_set(self, mock_futil, mock_elf,
                                    mock_rpath, mock_isvol, mock_fok, mock_sel,
                                    mock_getvol):
        """Test07 FakechrootEngine._fakechroot_env_set()."""
        Config.conf['fakechroot_expand_symlinks'] = None
        mock_getvol.return_value = ('v1:v2', 'm1:m2')
        mock_sel.return_value = '/ROOT/lib/libfakechroot.so'
        mock_fok.return_value = "/sys/infin"
        mock_rpath.return_value = '/bin/fake.so'
        mock_elf.return_value.get_ld_library_path.return_value = '/a'
        mock_elf.return_value.get_container_loader.return_value = '/ROOT/elf'
        mock_futil.return_value.find_file_in_dir.return_value = True
        mock_isvol.return_value = False
        self.xmode.get_mode.return_value = 'F1'
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake._elfpatcher = mock_elf
        ufake._fakechroot_env_set()
        self.assertTrue(self.xmode.get_mode.called)
        self.assertTrue(mock_getvol.called)
        self.assertTrue(mock_sel.called)
        self.assertTrue(mock_fok.called)
        self.assertTrue(mock_rpath.called)
        self.assertTrue(mock_isvol.called)

        Config.conf['fakechroot_expand_symlinks'] = None
        mock_getvol.return_value = ('v1:v2', 'm1:m2')
        mock_sel.return_value = '/ROOT/lib/libfakechroot.so'
        mock_fok.return_value = "/sys/infin"
        mock_rpath.return_value = '/bin/fake.so'
        mock_elf.return_value.get_ld_library_path.return_value = '/a'
        mock_elf.return_value.select_patchelf.return_value = '/ROOT/elf'
        mock_elf.return_value.get_patch_last_time.return_value = '/ROOT/elf'
        mock_elf.return_value.get_container_loader.return_value = '/ROOT/elf'
        mock_futil.return_value.find_file_in_dir.return_value = True
        mock_isvol.return_value = False
        self.xmode.get_mode.return_value = 'F4'
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake._elfpatcher = mock_elf
        ufake._fakechroot_env_set()
        self.assertTrue(mock_getvol.called)
        self.assertTrue(mock_sel.called)
        self.assertTrue(mock_fok.called)
        self.assertTrue(mock_rpath.called)
        self.assertTrue(mock_isvol.called)

    @patch('udocker.engine.fakechroot.Msg')
    def test_08__run_invalid_options(self, mock_msg):
        """Test08 FakechrootEngine._run_invalid_options()"""
        mock_msg.level = 5
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake.opt['netcoop'] = False
        ufake.opt['portsmap'] = True
        ufake._run_invalid_options()
        self.assertTrue(mock_msg.called)

    @patch('udocker.engine.fakechroot.FileUtil.get1stline')
    @patch('udocker.engine.fakechroot.FileUtil.cont2host')
    @patch('udocker.engine.fakechroot.FileUtil.find_exec')
    @patch('udocker.engine.fakechroot.OSInfo.get_filetype')
    @patch('udocker.engine.fakechroot.Msg')
    def test_09__run_add_script_support(self, mock_msg, mock_ftype,
                                        mock_findexe, mock_cont2host,
                                        mock_1stline):
        """Test09 FakechrootEngine._run_add_script_support()"""
        mock_msg.level = 3
        mock_ftype.return_value = "/bin/ls: ELF, x86-64, static"
        ufake = FakechrootEngine(self.local, self.xmode)
        status = ufake._run_add_script_support("/bin/ls")
        self.assertEqual(status, [])

        mock_msg.level = 3
        mock_ftype.return_value = "/bin/ls: xxx, x86-64, yyy"
        mock_findexe.side_effect = ["ls", ""]
        ufake = FakechrootEngine(self.local, self.xmode)
        status = ufake._run_add_script_support("/bin/ls")
        self.assertEqual(status, ["/ls"])

        mock_msg.level = 3
        mock_ftype.return_value = "/bin/ls: xxx, x86-64, yyy"
        mock_findexe.side_effect = ["", "ls"]
        mock_cont2host.return_value = "/bin/ls"
        mock_1stline.return_value = ""
        ufake = FakechrootEngine(self.local, self.xmode)
        ufake.container_root = "/ROOT"
        status = ufake._run_add_script_support("/bin/ls")
        self.assertEqual(status, ["/ROOT/ls"])

    @patch('udocker.engine.fakechroot.FileUtil.cont2host')
    @patch.object(FakechrootEngine, '_run_banner')
    @patch.object(FakechrootEngine, '_run_add_script_support')
    @patch.object(FakechrootEngine, '_set_cpu_affinity')
    @patch.object(FakechrootEngine, '_run_env_cleanup_list')
    @patch.object(FakechrootEngine, '_fakechroot_env_set')
    @patch.object(FakechrootEngine, '_run_env_set')
    @patch.object(FakechrootEngine, '_run_invalid_options')
    @patch.object(FakechrootEngine, '_run_init')
    @patch.object(FakechrootEngine, '_uid_check')
    @patch('udocker.engine.fakechroot.subprocess.call')
    @patch('udocker.engine.fakechroot.ElfPatcher')
    @patch('udocker.engine.fakechroot.Msg')
    def test_10_run(self, mock_msg, mock_elfp, mock_call, mock_uidc,
                    mock_rinit, mock_rinval, mock_renv, mock_fakeenv,
                    mock_renvclean, mock_setaff, mock_raddsup,
                    mock_runban, mock_cont2host):
        """Test10 FakechrootEngine.run()."""
        mock_uidc.return_value = None
        mock_rinit.return_value = ""
        ufake = FakechrootEngine(self.local, self.xmode)
        status = ufake.run("12345")
        self.assertEqual(status, 2)


        mock_uidc.return_value = None
        mock_rinit.return_value = '/bin/exec'
        mock_rinval.return_value = None
        self.xmode.get_mode.return_value = 'F1'
        mock_elfp.return_value.check_container_path.return_value = False
        mock_renv.return_value = None
        mock_fakeenv.return_value = None
        mock_renvclean.return_value = None
        mock_setaff.return_value = ['1', '2']
        mock_elfp.return_value.get_container_loader.return_value = '/ROOT/xx'
        mock_raddsup.return_value = ['/ROOT/xx.sh']
        mock_runban.return_value = None
        mock_cont2host.return_value = "/cont/ROOT"
        mock_call.return_value = 0
        ufake = FakechrootEngine(self.local, self.xmode)
        status = ufake.run("12345")
        self.assertEqual(status, 0)
        self.assertTrue(mock_uidc.called)
        self.assertTrue(mock_rinit.called)
        self.assertTrue(mock_msg.called)
        self.assertTrue(mock_call.called)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Alternative implementations for subprocess"""

import os
import sys
import subprocess

from udocker import is_genstr
from udocker.msg import Msg
from udocker.config import Config


class Uprocess(object):
    """Provide alternative implementations for subprocess"""

    def find_inpath(self, filename, path, rootdir=""):
        """Find file in a path set such as PATH=/usr/bin:/bin"""
        if not (filename and path):
            return ""
        basename = os.path.basename(filename)
        if is_genstr(path):
            if "=" in path:
                path = "".join(path.split("=", 1)[1:])
            path = path.split(":")
        if isinstance(path, (list, tuple)):
            for directory in path:
                full_path = rootdir + directory + "/" + basename
                if os.path.lexists(full_path):
                    return directory + "/" + basename
            return ""
        return ""

    def _check_output(self, *popenargs, **kwargs):
        """Alternative to subprocess.check_output"""
        process = subprocess.Popen(*popenargs, stdout=subprocess.PIPE, **kwargs)
        output, dummy = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd)
        return output

    def check_output(self, *popenargs, **kwargs):
        """Select check_output implementation"""
        try:
            # if Python 3
            if sys.version_info[0] >= 3:
                output = subprocess.check_output(*popenargs, **kwargs)
                chk_out = output.decode()
            # if Python >= 2.7
            elif sys.version_info[0] >= 2 and sys.version_info[1] >= 7:
                chk_out = subprocess.check_output(*popenargs, **kwargs)
            # if Python < 2.7
            else:
                chk_out = self._check_output(*popenargs, **kwargs)
        except OSError:
            return ""

        return chk_out

    def get_output(self, cmd, ignore_error=False):
        """Execute a shell command and get its output"""
        if not cmd[0].startswith("/"):
            path = Config.conf["root_path"] + ":" + os.getenv("PATH", "")
            cmd_path = self.find_inpath(cmd[0], path)
            if cmd_path:
                cmd[0] = cmd_path
        content = ""
        try:
            content = self.check_output(cmd, shell=False, stderr=Msg.chlderr,
                                        close_fds=True)
        except subprocess.CalledProcessError:
            if not ignore_error:
                return None
        return content.strip()

    def call(self, cmd, **kwargs):
        """Execute one shell command"""
        if not cmd[0].startswith("/"):
            path = Config.conf["root_path"] + ":" + os.getenv("PATH", "")
            cmd[0] = self.find_inpath(cmd[0], path)
        kwargs["shell"] = False
        return subprocess.call(cmd, **kwargs)

    def pipe(self, cmd1, cmd2, **kwargs):
        """Pipe two shell commands"""
        path = Config.conf["root_path"] + ":" + os.getenv("PATH", "")
        if not cmd1[0].startswith("/"):
            cmd1[0] = self.find_inpath(cmd1[0], path)
        if not cmd2[0].startswith("/"):
            cmd2[0] = self.find_inpath(cmd2[0], path)
        try:
            proc_1 = subprocess.Popen(cmd1, stderr=Msg.chlderr, shell=False,
                                      stdout=subprocess.PIPE, **kwargs)
        except (OSError, ValueError):
            return False
        try:
            proc_2 = subprocess.Popen(cmd2, stderr=Msg.chlderr, shell=False,
                                      stdin=proc_1.stdout)
        except (OSError, ValueError):
            proc_1.kill()
            return False
        while proc_1.returncode is None or proc_2.returncode is None:
            proc_1.wait()
            proc_2.wait()
        return not (proc_1.returncode or proc_2.returncode)



#!/usr/bin/env python
"""
udocker unit tests: NixAuthentication
"""

import pwd
import grp
from unittest import TestCase, main
from unittest.mock import patch, mock_open
from io import StringIO
from udocker.helper.nixauth import NixAuthentication
from udocker.config import Config

BUILTINS = "builtins"


class NixAuthenticationTestCase(TestCase):
    """Test NixAuthentication() *nix authentication portably."""

    def setUp(self):
        Config().getconf()

    def tearDown(self):
        pass

    def test_01_init(self):
        """Test01 NixAuthentication() constructor."""
        auth = NixAuthentication()
        self.assertEqual(auth.passwd_file, None)
        self.assertEqual(auth.group_file, None)

        auth = NixAuthentication("passwd", "group")
        self.assertEqual(auth.passwd_file, "passwd")
        self.assertEqual(auth.group_file, "group")

    # def test_02__user_in_subid(self):
    #     """Test02 NixAuthentication()._user_in_subid."""

    @patch.object(NixAuthentication, '_get_user_from_host')
    @patch.object(NixAuthentication, '_get_user_from_file')
    def test_03_user_in_subuid(self, mock_file, mock_host):
        """Test03 NixAuthentication().user_in_subuid."""
        mock_host.return_value = ("user", "1000", "1000", "usr",
                                  "/home/user", "/bin/bash")
        auth = NixAuthentication()
        auth.subuid_file = "/etc/subuid"
        subuid_line = StringIO('user:100000:65536')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(subuid_line.readline, ''))
            listuser = auth.user_in_subuid("user")
            self.assertEqual(listuser, [("100000", "65536")])
            self.assertTrue(mock_host.called)

        mock_file.return_value = ("user", "1000", "1000", "usr",
                                  "/home/user", "/bin/bash")
        auth = NixAuthentication()
        auth.passwd_file = "/etc/passwd"
        auth.subuid_file = "/etc/subuid"
        subuid_line = StringIO('user:100000:65536')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(subuid_line.readline, ''))
            listuser = auth.user_in_subuid("user")
            self.assertEqual(listuser, [("100000", "65536")])
            self.assertTrue(mock_file.called)

    @patch.object(NixAuthentication, '_get_user_from_host')
    @patch.object(NixAuthentication, '_get_user_from_file')
    def test_04_user_in_subgid(self, mock_file, mock_host):
        """Test04 NixAuthentication().user_in_subgid."""
        mock_host.return_value = ("user", "1000", "1000", "usr",
                                  "/home/user", "/bin/bash")
        auth = NixAuthentication()
        auth.subuid_file = "/etc/subgid"
        subgid_line = StringIO('user:100000:65536')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(subgid_line.readline, ''))
            listuser = auth.user_in_subgid("user")
            self.assertEqual(listuser, [("100000", "65536")])
            self.assertTrue(mock_host.called)

        mock_file.return_value = ("user", "1000", "1000", "usr",
                                  "/home/user", "/bin/bash")
        auth = NixAuthentication()
        auth.passwd_file = "/etc/passwd"
        auth.subgid_file = "/etc/subgid"
        subgid_line = StringIO('user:100000:65536')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(subgid_line.readline, ''))
            listuser = auth.user_in_subgid("user")
            self.assertEqual(listuser, [("100000", "65536")])
            self.assertTrue(mock_file.called)

    @patch('udocker.helper.nixauth.pwd.getpwnam')
    @patch('udocker.helper.nixauth.pwd.getpwuid')
    def test_05__get_user_from_host(self, mock_getpwuid, mock_getpwnam):
        """Test05 NixAuthentication()._get_user_from_host()."""
        usr = pwd.struct_passwd(["root", "*", "0", "0", "root usr",
                                 "/root", "/bin/bash"])
        mock_getpwuid.return_value = usr
        auth = NixAuthentication()
        (name, uid, gid, gecos, _dir, shell) = auth._get_user_from_host(0)
        self.assertEqual(name, usr.pw_name)
        self.assertEqual(uid, usr.pw_uid)
        self.assertEqual(gid, str(usr.pw_gid))
        self.assertEqual(gecos, usr.pw_gecos)
        self.assertEqual(_dir, usr.pw_dir)
        self.assertEqual(shell, usr.pw_shell)

        mock_getpwnam.return_value = usr
        auth = NixAuthentication()
        (name, uid, gid, gecos, _dir, shell) = auth._get_user_from_host("root")
        self.assertEqual(name, usr.pw_name)
        self.assertEqual(uid, usr.pw_uid)
        self.assertEqual(gid, str(usr.pw_gid))
        self.assertEqual(gecos, usr.pw_gecos)
        self.assertEqual(_dir, usr.pw_dir)

    # TODO: (mdavid) review test
    @patch('udocker.helper.nixauth.grp.getgrnam')
    @patch('udocker.helper.nixauth.grp.getgrgid')
    def test_06__get_group_from_host(self, mock_grgid, mock_grname):
        """Test06 NixAuthentication()._get_group_from_host()."""
        hgr = grp.struct_group(["root", "*", "0", str([])])
        mock_grgid.return_value = hgr
        auth = NixAuthentication()
        (name, gid, mem) = auth._get_group_from_host(0)
        self.assertEqual(name, hgr.gr_name)
        self.assertEqual(gid, str(hgr.gr_gid))
        self.assertEqual(mem, hgr.gr_mem)

        mock_grname.return_value = hgr
        auth = NixAuthentication()
        (name, gid, mem) = auth._get_group_from_host("root")
        self.assertEqual(name, hgr.gr_name)
        self.assertEqual(gid, str(hgr.gr_gid))
        self.assertEqual(mem, hgr.gr_mem)

    def test_07__get_user_from_file(self):
        """Test07 NixAuthentication()._get_user_from_file()."""
        auth = NixAuthentication()
        auth.passwd_file = "passwd"
        passwd_line = StringIO('root:x:0:0:root:/root:/bin/bash')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(passwd_line.readline, ''))
            (name, uid, gid,
             gecos, _dir, shell) = auth._get_user_from_file("root")
            self.assertEqual(name, "root")
            self.assertEqual(uid, "0")
            self.assertEqual(gid, "0")
            self.assertEqual(gecos, "root")
            self.assertEqual(_dir, "/root")
            self.assertEqual(shell, "/bin/bash")

        passwd_line = StringIO('root:x:0:0:root:/root:/bin/bash')
        auth = NixAuthentication()
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(passwd_line.readline, ''))
            (name, uid, gid,
             gecos, _dir, shell) = auth._get_user_from_file(0)
            self.assertEqual(name, "root")
            self.assertEqual(uid, "0")
            self.assertEqual(gid, "0")
            self.assertEqual(gecos, "root")
            self.assertEqual(_dir, "/root")
            self.assertEqual(shell, "/bin/bash")

    def test_08__get_group_from_file(self):
        """Test08 NixAuthentication()._get_group_from_file()."""
        auth = NixAuthentication()
        auth.passwd_file = "passwd"
        group_line = StringIO('root:x:0:a,b,c')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(group_line.readline, ''))
            (name, gid, mem) = auth._get_group_from_file("root")
            self.assertEqual(name, "root")
            self.assertEqual(gid, "0")
            self.assertEqual(mem, "a,b,c")

        group_line = StringIO('root:x:0:a,b,c')
        auth = NixAuthentication()
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(group_line.readline, ''))
            (name, gid, mem) = auth._get_group_from_file(0)
            self.assertEqual(name, "root")
            self.assertEqual(gid, "0")
            self.assertEqual(mem, "a,b,c")

    def test_09_add_user(self):
        """Test09 NixAuthentication().add_user()."""
        auth = NixAuthentication()
        with patch(BUILTINS + '.open', mock_open()):
            status = auth.add_user("root", "pw", 0, 0, "gecos",
                                   "/home", "/bin/bash")
            self.assertTrue(status)

    def test_10_add_group(self):
        """Test10 NixAuthentication().add_group()."""
        auth = NixAuthentication()
        with patch(BUILTINS + '.open', mock_open()):
            status = auth.add_group("root", 0)
            self.assertTrue(status)

    @patch.object(NixAuthentication, '_get_user_from_host')
    @patch.object(NixAuthentication, '_get_user_from_file')
    def test_11_get_user(self, mock_file, mock_host):
        """Test11 NixAuthentication().get_user()."""
        auth = NixAuthentication()
        auth.passwd_file = ""
        auth.get_user("user")
        self.assertTrue(mock_host.called)
        self.assertFalse(mock_file.called)

        mock_host.reset_mock()
        mock_file.reset_mock()
        auth = NixAuthentication()
        auth.passwd_file = "passwd"
        auth.get_user("user")
        self.assertFalse(mock_host.called)
        self.assertTrue(mock_file.called)

    @patch.object(NixAuthentication, '_get_group_from_host')
    @patch.object(NixAuthentication, '_get_group_from_file')
    def test_12_get_group(self, mock_file, mock_host):
        """Test12 NixAuthentication().get_group()."""
        auth = NixAuthentication()
        auth.group_file = ""
        auth.get_group("group")
        self.assertTrue(mock_host.called)
        self.assertFalse(mock_file.called)

        mock_host.reset_mock()
        mock_file.reset_mock()
        auth = NixAuthentication()
        auth.group_file = "group"
        auth.get_group("group")
        self.assertFalse(mock_host.called)
        self.assertTrue(mock_file.called)

    @patch.object(NixAuthentication, 'get_user')
    def test_13_get_home(self, mock_user):
        """Test13 NixAuthentication().get_home()."""
        mock_user.return_value = ('root', '0', '0', 'root',
                                  '/root', '/bin/bash')
        auth = NixAuthentication()
        status = auth.get_home()
        self.assertEqual(status, '/root')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Common Class for LocalFile API"""

import os
import time

from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil
from udocker.utils.uprocess import Uprocess
from udocker.helper.unique import Unique
from udocker.helper.hostinfo import HostInfo
from udocker.container.structure import ContainerStructure
from udocker.engine.execmode import ExecutionMode


class CommonLocalFileApi(object):
    """Common methods for Docker and OCI files"""

    def __init__(self, localrepo):
        self.localrepo = localrepo
        self._imagerepo = None

    def _move_layer_to_v1repo(self, filepath, layer_id, linkname=None):
        """Copy or rename an image layer file to a v1 repository"""
        if filepath.endswith("json"):
            target_file = self.localrepo.layersdir + '/' + layer_id + ".json"
        elif filepath.endswith("layer.tar"):
            target_file = self.localrepo.layersdir + '/' + layer_id + ".layer"
        elif ':' in layer_id:
            (dummy, layer_id) = layer_id.split(':', 1)
            target_file = self.localrepo.layersdir + '/' + layer_id + ".layer"
        else:
            return False
        try:
            os.rename(filepath, target_file)
        except(IOError, OSError):
            if not FileUtil(filepath).copyto(target_file):
                return False
        self.localrepo.add_image_layer(target_file, linkname)
        return True

    def _load_image_step2(self, structure, imagerepo, tag):
        """Implementation specific image load"""

    def _load_image(self, structure, imagerepo, tag):
        """Load a container image into a repository mimic docker load"""
        if self._imagerepo and self._imagerepo != imagerepo:
            cd_imagerepo = self._imagerepo
        else:
            cd_imagerepo = imagerepo

        if self.localrepo.cd_imagerepo(cd_imagerepo, tag):
            Msg().err("Error: repository and tag already exist",
                      cd_imagerepo, tag)
            return []

        self.localrepo.setup_imagerepo(cd_imagerepo)
        tag_dir = self.localrepo.setup_tag(tag)
        if not tag_dir:
            Msg().err("Error: setting up repository", cd_imagerepo, tag)
            return []

        if not self.localrepo.set_version("v1"):
            Msg().err("Error: setting repository version")
            return []

        return self._load_image_step2(structure, imagerepo, tag)

    def _untar_saved_container(self, tarfile, destdir):
        """Untar container created with docker save"""
        #umask 022
        verbose = ''
        if Msg.level >= Msg.VER:
            verbose = 'v'

        cmd = ["tar", "-C", destdir, "-x" + verbose,
               "--delay-directory-restore", "--one-file-system",
               "--no-same-owner", "--no-same-permissions", "--overwrite",
               "-f", tarfile]

        status = Uprocess().call(cmd, stderr=Msg.chlderr, close_fds=True)
        return not status

    def create_container_meta(self, layer_id, comment="created by udocker"):
        """Create metadata for a given container layer, used in import.
        A file for import is a tarball of a directory tree, does not contain
        metadata. This method creates minimal metadata.
        """
        container_json = {}
        container_json["id"] = layer_id
        container_json["comment"] = comment
        container_json["created"] = \
            time.strftime("%Y-%m-%dT%H:%M:%S.000000000Z")
        container_json["architecture"] = HostInfo().arch()
        container_json["os"] = HostInfo().osversion()
        layer_file = self.localrepo.layersdir + '/' + layer_id + ".layer"
        container_json["size"] = FileUtil(layer_file).size()
        if container_json["size"] == -1:
            container_json["size"] = 0
        container_json["container_config"] = {
            "Hostname": "",
            "Domainname": "",
            "User": "",
            "Memory": 0,
            "MemorySwap": 0,
            "CpusShares": 0,
            "Cpuset": "",
            "AttachStdin": False,
            "AttachStdout": False,
            "AttachStderr": False,
            "PortSpecs": None,
            "ExposedPorts": None,
            "Tty": False,
            "OpenStdin": False,
            "StdinOnce": False,
            "Env": None,
            "Cmd": None,
            "Image": "",
            "Volumes": None,
            "WorkingDir": "",
            "Entrypoint": None,
            "NetworkDisable": False,
            "MacAddress": "",
            "OnBuild": None,
            "Labels": None
        }
        container_json["config"] = {
            "Hostname": "",
            "Domainname": "",
            "User": "",
            "Memory": 0,
            "MemorySwap": 0,
            "CpusShares": 0,
            "Cpuset": "",
            "AttachStdin": False,
            "AttachStdout": False,
            "AttachStderr": False,
            "PortSpecs": None,
            "ExposedPorts": None,
            "Tty": False,
            "OpenStdin": False,
            "StdinOnce": False,
            "Env": None,
            "Cmd": None,
            "Image": "",
            "Volumes": None,
            "WorkingDir": "",
            "Entrypoint": None,
            "NetworkDisable": False,
            "MacAddress": "",
            "OnBuild": None,
            "Labels": None
        }
        return container_json

    def import_toimage(self, tarfile, imagerepo, tag, move_tarball=True):
        """Import a tar file containing a simple directory tree possibly
        created with Docker export and create local image"""
        if not os.path.exists(tarfile) and tarfile != '-':
            Msg().err("Error: tar file does not exist: ", tarfile)
            return False
        self.localrepo.setup_imagerepo(imagerepo)
        tag_dir = self.localrepo.cd_imagerepo(imagerepo, tag)
        if tag_dir:
            Msg().err("Error: tag already exists in repo:", tag)
            return False
        tag_dir = self.localrepo.setup_tag(tag)
        if not tag_dir:
            Msg().err("Error: creating repo and tag")
            return False
        if not self.localrepo.set_version("v1"):
            Msg().err("Error: setting repository version")
            return False
        layer_id = Unique().layer_v1()
        layer_file = self.localrepo.layersdir + '/' + layer_id + ".layer"
        json_file = self.localrepo.layersdir + '/' + layer_id + ".json"
        if move_tarball:
            try:
                os.rename(tarfile, layer_file)
            except(IOError, OSError):
                pass
        if not os.path.exists(layer_file):
            if not FileUtil(tarfile).copyto(layer_file):
                Msg().err("Error: in move/copy file", tarfile)
                return False
        self.localrepo.add_image_layer(layer_file)
        self.localrepo.save_json("ancestry", [layer_id])
        container_json = self.create_container_meta(layer_id)
        self.localrepo.save_json(json_file, container_json)
        self.localrepo.add_image_layer(json_file)
        Msg().out("Info: added layer", layer_id, l=Msg.INF)
        return layer_id

    def import_tocontainer(self, tarfile, imagerepo, tag, container_name):
        """Import a tar file containing a simple directory tree possibly
        created with Docker export and create local container ready to use"""
        if not imagerepo:
            imagerepo = "IMPORTED"
            tag = "latest"
        if not os.path.exists(tarfile) and tarfile != '-':
            Msg().err("Error: tar file does not exist:", tarfile)
            return False
        if container_name:
            if self.localrepo.get_container_id(container_name):
                Msg().err("Error: container name already exists:",
                          container_name)
                return False
        layer_id = Unique().layer_v1()
        container_json = self.create_container_meta(layer_id)
        container_id = ContainerStructure(self.localrepo).create_fromlayer(
            imagerepo, tag, tarfile, container_json)
        if container_name:
            self.localrepo.set_container_name(container_id, container_name)
        return container_id

    def import_clone(self, tarfile, container_name):
        """Import a tar file containing a clone of a udocker container
        created with export --clone and create local cloned container
        ready to use
        """
        if not os.path.exists(tarfile) and tarfile != '-':
            Msg().err("Error: tar file does not exist:", tarfile)
            return False
        if container_name:
            if self.localrepo.get_container_id(container_name):
                Msg().err("Error: container name already exists:",
                          container_name)
                return False
        container_id = ContainerStructure(self.localrepo).clone_fromfile(
            tarfile)
        if container_name:
            self.localrepo.set_container_name(container_id, container_name)
        return container_id

    def clone_container(self, container_id, container_name):
        """Clone/duplicate an existing container creating a complete
        copy including metadata, control files, and rootfs, The copy
        will have a new id.
        """
        if container_name:
            if self.localrepo.get_container_id(container_name):
                Msg().err("Error: container name already exists:",
                          container_name)
                return False
        dest_container_id = ContainerStructure(self.localrepo,
                                               container_id).clone()
        if container_name:
            self.localrepo.set_container_name(dest_container_id,
                                              container_name)
        exec_mode = ExecutionMode(self.localrepo, dest_container_id)
        xmode = exec_mode.get_mode()
        if xmode.startswith('F'):
            exec_mode.set_mode(xmode, True)
        return dest_container_id

    def _get_imagedir_type(self, tmp_imagedir):
        """Identify type of image from extracted image directory"""
        image_types_list = [(tmp_imagedir + "/oci-layout", "OCI"),
                            (tmp_imagedir + "/manifest.json", "Docker"), ]
        for (checkfile, imagetype) in image_types_list:
            if os.path.exists(checkfile):
                return imagetype
        return ""



# -*- coding: utf-8 -*-
"""PRoot execution engine"""

import sys
import os
import subprocess

from udocker.msg import Msg
from udocker.config import Config
from udocker.engine.base import ExecutionEngineCommon
from udocker.utils.fileutil import FileUtil
from udocker.utils.uvolume import Uvolume
from udocker.helper.hostinfo import HostInfo
from udocker.utils.uprocess import Uprocess


class PRootEngine(ExecutionEngineCommon):
    """Docker container execution engine using PRoot
    Provides a chroot like environment to run containers.
    Uses PRoot both as chroot alternative and as emulator of
    the root identity and privileges.
    Inherits from ContainerEngine class
    """

    def __init__(self, localrepo, exec_mode):
        super(PRootEngine, self).__init__(localrepo, exec_mode)
        self.executable = None                   # PRoot
        self.proot_noseccomp = False             # No seccomp mode
        self.proot_newseccomp = False            # New seccomp mode
        self._kernel = HostInfo().oskernel()     # Emulate kernel

    def select_proot(self):
        """Set proot executable and related variables"""
        self.executable = Config.conf['use_proot_executable']
        if self.executable != "UDOCKER" and not self.executable:
            self.executable = FileUtil("proot").find_exec()

        if self.executable == "UDOCKER" or not self.executable:
            self.executable = ""
            arch = HostInfo().arch()
            image_list = []
            if arch == "amd64":
                if HostInfo().oskernel_isgreater([4, 8, 0]):
                    image_list = ["proot-x86_64-4_8_0", "proot-x86_64", "proot"]
                else:
                    image_list = ["proot-x86_64", "proot"]
            elif arch == "i386":
                if HostInfo().oskernel_isgreater([4, 8, 0]):
                    image_list = ["proot-x86-4_8_0", "proot-x86", "proot"]
                else:
                    image_list = ["proot-x86", "proot"]
            elif arch == "arm64":
                if HostInfo().oskernel_isgreater([4, 8, 0]):
                    image_list = ["proot-arm64-4_8_0", "proot-arm64", "proot"]
                else:
                    image_list = ["proot-arm64", "proot"]
            elif arch == "arm":
                if HostInfo().oskernel_isgreater([4, 8, 0]):
                    image_list = ["proot-arm-4_8_0", "proot-arm", "proot"]
                else:
                    image_list = ["proot-arm", "proot"]
            f_util = FileUtil(self.localrepo.bindir)
            self.executable = f_util.find_file_in_dir(image_list)

        if not os.path.exists(self.executable):
            Msg().err("Error: proot executable not found")
            sys.exit(1)

        if Config.conf['proot_noseccomp'] is not None:
            self.proot_noseccomp = Config.conf['proot_noseccomp']
        if self.exec_mode.get_mode() == "P2":
            self.proot_noseccomp = True
        if self._is_seccomp_patched(self.executable):
            self.proot_newseccomp = True

    def _is_seccomp_patched(self, executable):
        """Check if kernel has ptrace/seccomp fixes added
           on 4.8.0.
           Only required for kernels below 4.8.0 to check
           if the patch has been backported e.g CentOS 7
        """
        if "PROOT_NEW_SECCOMP" in os.environ:
            return True
        if ("PROOT_NO_SECCOMP" in os.environ or
                self.proot_noseccomp or
                HostInfo().oskernel_isgreater([4, 8, 0])):
            return False
        host_file = self.container_dir + "/osenv.json"
        host_info = self._is_same_osenv(host_file)
        if host_info:
            if "PROOT_NEW_SECCOMP" in host_info:
                return True
            return False
        out = Uprocess().get_output([executable, "-r", "/",
                                     executable, "--help"])
        if not out:
            os.environ["PROOT_NEW_SECCOMP"] = "1"
            out = Uprocess().get_output([executable, "-r", "/",
                                         executable, "--help"])
            del os.environ["PROOT_NEW_SECCOMP"]
            if out:
                self._save_osenv(host_file,
                                      dict([("PROOT_NEW_SECCOMP", 1), ]))
                return True
        self._save_osenv(host_file)
        return False

    def _set_uid_map(self):
        """Set the uid_map string for container run command"""
        if self.opt["uid"] == "0":
            uid_map_list = ["-0", ]
        else:
            uid_map_list = \
                ["-i", self.opt["uid"] + ":" + self.opt["gid"], ]
        return uid_map_list

    def _create_mountpoint(self, host_path, cont_path, dirs_only=False):
        """Override create mountpoint"""
        return True

    def _get_volume_bindings(self):
        """Get the volume bindings string for container run command"""
        proot_vol_list = []
        for vol in self.opt["vol"]:
            proot_vol_list.extend(["-b", "%s:%s" % Uvolume(vol).split()])
        return proot_vol_list

    def _get_network_map(self):
        """Get mapping of TCP/IP ports"""
        proot_netmap_list = []
        for (cont_port, host_port) in list(self._get_portsmap().items()):
            proot_netmap_list.extend(["-p", "%d:%d" % (cont_port, host_port)])
        if self.opt["netcoop"]:
            proot_netmap_list.extend(["-n", ])
        return proot_netmap_list

    def run(self, container_id):
        """Execute a Docker container using PRoot. This is the main method
        invoked to run the a container with PRoot.
          * argument: container_id or name
          * options:  many via self.opt see the help
        """

        # setup execution
        if not self._run_init(container_id):
            return 2

        self.select_proot()

        # seccomp and ptrace behavior change on 4.8.0 onwards
        if self.proot_noseccomp or os.getenv("PROOT_NO_SECCOMP"):
            self.opt["env"].append("PROOT_NO_SECCOMP=1")

        if self.proot_newseccomp or os.getenv("PROOT_NEW_SECCOMP"):
            self.opt["env"].append("PROOT_NEW_SECCOMP=1")

        if not HostInfo().oskernel_isgreater([3, 0, 0]):
            self._kernel = "6.0.0"

        if self.opt["kernel"]:
            self._kernel = self.opt["kernel"]

        # set environment variables
        self._run_env_set()

        if Msg.level >= Msg.DBG:
            proot_verbose = ["-v", "9", ]
        else:
            proot_verbose = []

        if (Config.conf['proot_killonexit'] and
                self._has_option("--kill-on-exit")):
            proot_kill_on_exit = ["--kill-on-exit", ]
        else:
            proot_kill_on_exit = []

        # build the actual command
        cmd_l = self._set_cpu_affinity()
        cmd_l.append(self.executable)
        cmd_l.extend(proot_verbose)
        cmd_l.extend(proot_kill_on_exit)
        cmd_l.extend(self._get_volume_bindings())
        cmd_l.extend(self._set_uid_map())
        cmd_l.extend(["-k", self._kernel, ])
        cmd_l.extend(self._get_network_map())
        cmd_l.extend(["-r", self.container_root, ])

        if self.opt["cwd"]:  # set current working directory
            cmd_l.extend(["-w", self.opt["cwd"], ])
        cmd_l.extend(self.opt["cmd"])
        Msg().out("CMD =", cmd_l, l=Msg.VER)

        # cleanup the environment
        self._run_env_cleanup_dict()

        # execute
        self._run_banner(self.opt["cmd"][0])
        status = subprocess.call(cmd_l, shell=False, close_fds=True,
                                 env=os.environ.update(self.opt["env"].dict()))
        return status



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
========
udocker
========
Wrapper to execute basic docker containers without using docker.
This tool is a last resort for the execution of docker containers
where docker is unavailable. It only provides a limited set of
functionalities.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import sys

sys.path.append(os.path.dirname(os.path.abspath(sys.argv[0])) + '/../')

# pylint: disable=wrong-import-position
from udocker.msg import Msg
from udocker.umain import UMain
from udocker.utils.fileutil import FileUtil


def main():
    """Program start and exception handling"""

    exit_status = 0
    try:
        exit_status = UMain(sys.argv).execute()
    except KeyboardInterrupt:
        Msg().out("Info: keyboard interrupt", l=Msg.INF)
        FileUtil().cleanup()
        exit_status = 1
    except SystemExit as error:
        FileUtil().cleanup()
        try:
            exit_status = int(error.args[0])
        except (ValueError, TypeError):
            exit_status = 1
    except:
        FileUtil().cleanup()
        exit_status = 1
        raise
    else:
        FileUtil().cleanup()

    sys.exit(exit_status)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Checksumming for files"""

import re

from udocker.utils.uprocess import Uprocess

try:
    import hashlib
except ImportError:
    pass

class ChkSUM(object):
    """Checksumming for files"""

    def __init__(self):
        self._algorithms = {}
        try:
            dummy = hashlib.sha256()
            self._algorithms["sha256"] = self._hashlib_sha256
        except NameError:
            self._algorithms["sha256"] = self._openssl_sha256
        try:
            dummy = hashlib.sha512()
            self._algorithms["sha512"] = self._hashlib_sha512
        except NameError:
            self._algorithms["sha512"] = self._openssl_sha512

    def _hashlib(self, algorithm, filename):
        """hash calculation using hashlib"""
        try:
            with open(filename, "rb") as filep:
                for chunk in iter(lambda: filep.read(4096), b""):
                    algorithm.update(chunk)
            return algorithm.hexdigest()
        except (IOError, OSError):
            return ""

    def _hashlib_sha256(self, filename):
        """sha256 calculation using hashlib"""
        return self._hashlib(hashlib.sha256(), filename)

    def _hashlib_sha512(self, filename):
        """sha512 calculation using hashlib"""
        return self._hashlib(hashlib.sha512(), filename)

    def _openssl(self, algorithm, filename):
        """hash calculation using openssl"""
        cmd = ["openssl", "dgst", "-hex", "-r", algorithm, filename]
        output = Uprocess().get_output(cmd)
        if output is None:
            return ""
        match = re.match("^(\\S+) ", output)
        if match:
            return match.group(1)
        return ""

    def _openssl_sha256(self, filename):
        """sha256 calculation using openssl"""
        return self._openssl("-sha256", filename)

    def _openssl_sha512(self, filename):
        """sha512 calculation using openssl"""
        return self._openssl("-sha512", filename)

    def sha256(self, filename):
        """Call the actual implementation selected in __init__"""
        return self._algorithms["sha256"](filename)

    def sha512(self, filename):
        """Call the actual implementation selected in __init__"""
        return self._algorithms["sha512"](filename)

    def hash(self, filename, algorithm):
        """Compute hash algorithm for file"""
        if algorithm in self._algorithms:
            return self._algorithms[algorithm](filename)
        return ""



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: GetURLexeCurl
"""

from unittest import TestCase, main
from unittest.mock import patch
from udocker.utils.curl import GetURLexeCurl
from udocker.config import Config


class GetURLexeCurlTestCase(TestCase):
    """GetURLexeCurl TestCase."""

    def setUp(self):
        Config().getconf()
        Config().conf['timeout'] = 1
        Config().conf['ctimeout'] = 1
        Config().conf['download_timeout'] = 1
        Config().conf['http_agent'] = ""
        Config().conf['http_proxy'] = ""

    def tearDown(self):
        pass

    def _get(self, *args, **kwargs):
        """Mock for pycurl.get."""
        return args[0]

    def test_01_init(self):
        """Test01 GetURLexeCurl() constructor."""
        geturl = GetURLexeCurl()
        self.assertIsNone(geturl._opts)
        self.assertIsNone(geturl._files)

    @patch('udocker.utils.curl.FileUtil.find_exec')
    def test_02_is_available(self, mock_findexec):
        """Test02 GetURLexeCurl()._is_available()."""
        mock_findexec.return_value = "/tmp"
        geturl = GetURLexeCurl()
        self.assertTrue(geturl.is_available())

        mock_findexec.return_value = ""
        geturl = GetURLexeCurl()
        self.assertFalse(geturl.is_available())

    # def test_03__select_implementation(self):
    #     """Test03 GetURLexeCurl()._select_implementation()."""

    @patch('udocker.utils.curl.FileUtil.mktmp')
    def test_04__set_defaults(self, mock_mktemp):
        """Test04 GetURLexeCurl()._set_defaults()"""
        mock_mktemp.side_effect = ["/tmp/err", "/tmp/out", "/tmp/hdr"]
        geturl = GetURLexeCurl()
        geturl._set_defaults()
        self.assertEqual(geturl._opts["insecure"], list())

        mock_mktemp.side_effect = ["/tmp/err", "/tmp/out", "/tmp/hdr"]
        geturl = GetURLexeCurl()
        geturl.insecure = True
        geturl._set_defaults()
        self.assertEqual(geturl._opts["insecure"], ["-k"])

        mock_mktemp.side_effect = ["/tmp/err", "/tmp/out", "/tmp/hdr"]
        geturl = GetURLexeCurl()
        geturl._set_defaults()
        self.assertEqual(geturl._files["url"], "")
        self.assertEqual(geturl._files["error_file"], "/tmp/err")
        self.assertEqual(geturl._files["header_file"], "/tmp/hdr")

    @patch('udocker.utils.curl.FileUtil.remove')
    @patch('udocker.utils.curl.json.dumps')
    def test_05__mkcurlcmd(self, mock_jdump, mock_furm):
        """Test05 GetURLexeCurl()._mkcurlcmd()."""
        argl = ["http://host"]
        kargl = list()
        Config.conf['use_curl_executable'] = ""
        mock_jdump.return_value = {"post": "pst1"}
        mock_furm.return_value = None
        res = ["curl", "-D", "/tmp/hdr", "-o", "/tmp/out",
               "--stderr", "/tmp/err", "['http://host']"]
        geturl = GetURLexeCurl()
        geturl._opts = dict()
        geturl._files = {"url": "", "error_file": "/tmp/err",
                         "output_file": "/tmp/out",
                         "header_file": "/tmp/hdr"}
        status = geturl._mkcurlcmd(argl, kargl)
        self.assertFalse(mock_jdump.called)
        self.assertEqual(status, res)

        # argl = ["http://host"]
        # kargl = {"post":  "pst1", "ctimeout": 1000,
        #          "timeout": 50, "proxy": "http://proxy",
        #          "header": ["Authorization: Bearer"], "v": True,
        #          "nobody": True, "resume": True,}
        # Config.conf['use_curl_executable'] = ""
        # mock_jdump.return_value = {"post": "pst1"}
        # mock_furm.return_value = None
        # res = ["curl", "-X", "POST", "-H",
        #        "Content-Type: application/json",
        #        "-d", "pst1", "--connect-timeout", "1000",
        #        "-m", "50", "--proxy", "http://proxy",
        #        "-H", "Authorization: Bearer", "-v",
        #        "-D", "/tmp/hdr", "-o", "/tmp/out",
        #        "--stderr", "/tmp/err", "['http://host']"]
        # geturl = GetURLexeCurl()
        # geturl._opts = dict()
        # geturl._files = {"url": "", "error_file": "/tmp/err",
        #                  "output_file": "/tmp/out",
        #                  "header_file": "/tmp/hdr"}
        # status = geturl._mkcurlcmd(argl, kargl)
        # self.assertFalse(mock_jdump.called)
        # self.assertEqual(status, res)

    def test_06_get(self):
        """Test06 GetURLexeCurl().get()."""
        geturl = GetURLexeCurl()
        geturl._geturl = type('test', (object,), {})()
        geturl.get = self._get
        self.assertEqual(geturl.get("http://host"), "http://host")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Implements most of the command line interface."""

import os
import sys
from udocker.msg import Msg
from udocker.cmdparser import CmdParser
from udocker.config import Config
from udocker.container.localrepo import LocalRepository
from udocker.cli import UdockerCLI


class UMain(object):
    """These methods correspond directly to the commands that can
    be invoked via the command line interface.
    """

    STATUS_OK = 0
    STATUS_ERROR = 1

    def __init__(self, argv):
        """Initialize variables of the class"""
        self.argv = argv
        self.cmdp = None
        self.local = None
        self.cli = None

    def _prepare_exec(self):
        """Prepare configuration, parse and execute the command line"""
        self.cmdp = CmdParser()
        self.cmdp.parse(self.argv)
        allow_root = self.cmdp.get("--allow-root", "GEN_OPT")
        if not (os.geteuid() or allow_root):
            Msg().err("Error: do not run as root !")
            sys.exit(self.STATUS_ERROR)

        if self.cmdp.get("--config=", "GEN_OPT"):
            conf_file = self.cmdp.get("--config=", "GEN_OPT")
            Config().getconf(conf_file)
        else:
            Config().getconf()

        if (self.cmdp.get("--debug", "GEN_OPT") or
                self.cmdp.get("-D", "GEN_OPT")):
            Config.conf['verbose_level'] = Msg.DBG
        elif (self.cmdp.get("--quiet", "GEN_OPT") or
              self.cmdp.get("-q", "GEN_OPT")):
            Config.conf['verbose_level'] = Msg.MSG
        Msg().setlevel(Config.conf['verbose_level'])

        if self.cmdp.get("--insecure", "GEN_OPT"):
            Config.conf['http_insecure'] = True

        topdir = self.cmdp.get("--repo=", "GEN_OPT")
        if topdir:  # override repo root tree
            Config.conf['topdir'] = topdir
        self.local = LocalRepository()
        if not self.local.is_repo():
            if topdir:
                Msg().err("Error: invalid udocker repository:", topdir)
                sys.exit(self.STATUS_ERROR)
            else:
                Msg().out("Info: creating repo: " + Config.conf['topdir'],
                          l=Msg.INF)
                self.local.create_repo()

        self.cli = UdockerCLI(self.local)

    # pylint: disable=too-many-return-statements
    def execute(self):
        """Command parsing and selection"""
        self._prepare_exec()
        cmds = {
            "search": self.cli.do_search, "help": self.cli.do_help,
            "images": self.cli.do_images, "pull": self.cli.do_pull,
            "create": self.cli.do_create, "ps": self.cli.do_ps,
            "run": self.cli.do_run, "version": self.cli.do_version,
            "rmi": self.cli.do_rmi, "mkrepo": self.cli.do_mkrepo,
            "import": self.cli.do_import, "load": self.cli.do_load,
            "export": self.cli.do_export, "clone": self.cli.do_clone,
            "protect": self.cli.do_protect, "rm": self.cli.do_rm,
            "name": self.cli.do_name, "rmname": self.cli.do_rmname,
            "verify": self.cli.do_verify, "logout": self.cli.do_logout,
            "unprotect": self.cli.do_unprotect, "rename": self.cli.do_rename,
            "showconf": self.cli.do_showconf, "save": self.cli.do_save,
            "inspect": self.cli.do_inspect, "login": self.cli.do_login,
            "setup": self.cli.do_setup, "install": self.cli.do_install,
        }

        if ((len(self.argv) == 1) or
                self.cmdp.get("-h", "GEN_OPT") or
                self.cmdp.get("--help", "GEN_OPT")):
            return self.cli.do_help(self.cmdp)

        if (self.cmdp.get("-V", "GEN_OPT") or
                self.cmdp.get("--version", "GEN_OPT")):
            return self.cli.do_version(self.cmdp)

        command = self.cmdp.get("", "CMD")
        if command in cmds:
            if self.cmdp.get("--help", "CMD_OPT"):
                Msg().out(cmds[command].__doc__)
                return self.STATUS_OK
            if command in ["version", "showconf"]:
                return cmds[command](self.cmdp)
            if command != "install":
                self.cli.do_install(None)
            exit_status = cmds[command](self.cmdp)  # executes command
            if self.cmdp.missing_options():
                Msg().err("Error: syntax error at: %s" %
                          " ".join(self.cmdp.missing_options()))
                return self.STATUS_ERROR

            return exit_status

        Msg().err("Error: invalid command:", command, "\n")
        return self.STATUS_ERROR



#!/usr/bin/env python
"""
udocker unit tests: RuncEngine
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch, mock_open
from udocker.config import Config
from udocker.engine.runc import RuncEngine

BUILTINS = "builtins"
BOPEN = BUILTINS + '.open'


class RuncEngineTestCase(TestCase):
    """Test RuncEngine() containers execution with runC."""

    def setUp(self):
        Config().getconf()
        Config().conf['hostauth_list'] = ("/etc/passwd", "/etc/group")
        Config().conf['cmd'] = "/bin/bash"
        Config().conf['cpu_affinity_exec_tools'] = \
            (["numactl", "-C", "%s", "--", ], ["taskset", "-c", "%s", ])
        Config().conf['runc_capabilities'] = [
            "CAP_KILL", "CAP_NET_BIND_SERVICE", "CAP_CHOWN", "CAP_DAC_OVERRIDE",
            "CAP_FOWNER", "CAP_FSETID", "CAP_KILL", "CAP_SETGID", "CAP_SETUID",
            "CAP_SETPCAP", "CAP_NET_BIND_SERVICE", "CAP_NET_RAW",
            "CAP_SYS_CHROOT", "CAP_MKNOD", "CAP_AUDIT_WRITE", "CAP_SETFCAP",
        ]
        Config().conf['valid_host_env'] = "HOME"
        Config().conf['username'] = "user"
        Config().conf['userhome'] = "/"
        Config().conf['oskernel'] = "4.8.13"
        Config().conf['location'] = ""
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

        str_exmode = 'udocker.engine.execmode.ExecutionMode'
        self.execmode = patch(str_exmode)
        self.xmode = self.execmode.start()
        self.mock_execmode = Mock()
        self.xmode.return_value = self.mock_execmode

    def tearDown(self):
        self.lrepo.stop()
        self.execmode.stop()

    def test_01_init(self):
        """Test01 RuncEngine() constructor."""
        rcex = RuncEngine(self.local, self.xmode)
        self.assertEqual(rcex.executable, None)
        self.assertEqual(rcex.execution_id, None)

    @patch('udocker.helper.elfpatcher.os.path.exists')
    @patch('udocker.engine.runc.HostInfo.arch')
    @patch('udocker.engine.runc.FileUtil.find_file_in_dir')
    @patch('udocker.engine.runc.FileUtil.find_exec')
    def test_02_select_runc(self, mock_findexe, mock_find,
                            mock_arch, mock_exists):
        """Test02 RuncEngine().select_runc()."""
        Config.conf['use_runc_executable'] = ""
        mock_findexe.return_value = "/bin/runc-arm"
        mock_exists.return_value = True
        rcex = RuncEngine(self.local, self.xmode)
        rcex.select_runc()
        self.assertTrue(mock_findexe.called)

        Config.conf['use_runc_executable'] = "UDOCKER"
        mock_arch.return_value = "amd64"
        mock_find.return_value = "runc-x86_64"
        mock_exists.return_value = True
        rcex = RuncEngine(self.local, self.xmode)
        rcex.select_runc()
        self.assertTrue(mock_arch.called)
        self.assertTrue(mock_find.called)

        Config.conf['use_runc_executable'] = "UDOCKER"
        mock_arch.return_value = "i386"
        mock_find.return_value = "runc-x86"
        mock_exists.return_value = True
        rcex = RuncEngine(self.local, self.xmode)
        rcex.select_runc()
        self.assertTrue(mock_arch.called)
        self.assertTrue(mock_find.called)

    @patch('udocker.engine.runc.json.load')
    @patch('udocker.engine.runc.FileUtil.register_prefix')
    @patch('udocker.engine.runc.FileUtil.remove')
    @patch('udocker.engine.runc.FileUtil.size')
    @patch('udocker.engine.runc.subprocess.call')
    @patch('udocker.engine.runc.os.path.realpath')
    def test_03__load_spec(self, mock_realpath, mock_call, mock_size,
                           mock_rm, mock_reg, mock_jload):
        """Test03 RuncEngine()._load_spec()."""
        mock_size.side_effect = [-1, -1]
        mock_realpath.return_value = "/container/ROOT"
        mock_call.return_value = 1
        mock_rm.return_value = None
        mock_reg.return_value = None
        rcex = RuncEngine(self.local, self.xmode)
        status = rcex._load_spec(False)
        self.assertFalse(mock_rm.called)
        self.assertFalse(mock_reg.called)
        self.assertTrue(mock_call.called)
        self.assertTrue(mock_realpath.called)
        self.assertFalse(status)

        jload = {"container": "cxxx", "parent": "dyyy",
                 "created": "2020-05-05T21:20:07.182447994Z",
                 "os": "linux",
                 "container_config": {"Tty": "false", "Cmd": ["/bin/sh"]},
                 "Image": "sha256:aa"
                }
        mock_size.side_effect = [100, 100]
        mock_realpath.return_value = "/container/ROOT"
        mock_call.return_value = 0
        mock_rm.return_value = None
        mock_reg.return_value = None
        mock_jload.return_value = jload
        with patch(BOPEN, mock_open()) as mopen:
            rcex = RuncEngine(self.local, self.xmode)
            status = rcex._load_spec(True)
            self.assertTrue(mopen.called)
            self.assertEqual(status, jload)
            self.assertTrue(mock_rm.called)
            self.assertTrue(mock_reg.called)

    @patch('udocker.engine.runc.json.dump')
    def test_04__save_spec(self, mock_jdump):
        """Test04 RuncEngine()._save_spec()."""
        jdump = {"container": "cxxx", "parent": "dyyy",
                 "created": "2020-05-05T21:20:07.182447994Z",
                 "os": "linux",
                 "container_config": {"Tty": "false", "Cmd": ["/bin/sh"]},
                 "Image": "sha256:aa"
                }
        mock_jdump.return_value = jdump
        with patch(BOPEN, mock_open()) as mopen:
            rcex = RuncEngine(self.local, self.xmode)
            status = rcex._save_spec()
            self.assertTrue(mopen.called)
            self.assertTrue(status)

    @patch('udocker.engine.runc.os.getgid')
    @patch('udocker.engine.runc.os.getuid')
    @patch('udocker.engine.runc.platform.node')
    @patch('udocker.engine.runc.os.path.realpath')
    def test_05__set_spec(self, mock_realpath, mock_node,
                          mock_getuid, mock_getgid):
        """Test05 RuncEngine()._set_spec()."""
        # rcex.opt["hostname"] has nodename
        mock_realpath.return_value = "/.udocker/containers/aaaaaa/ROOT"
        mock_node.return_value = "nodename.local"
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex._container_specjson = dict()
        rcex._container_specjson["root"] = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["linux"] = dict()
        rcex._container_specjson["linux"]["uidMappings"] = dict()
        rcex._container_specjson["linux"]["gidMappings"] = dict()
        rcex.opt["cwd"] = "/"
        rcex.opt["env"] = []
        rcex.opt["cmd"] = "bash"
        rcex.opt["hostname"] = "node.domain"
        json_obj = rcex._set_spec()
        self.assertEqual(json_obj["hostname"], "node.domain")

        # empty rcex.opt["hostname"]
        mock_realpath.return_value = "/.udocker/containers/aaaaaa/ROOT"
        mock_node.return_value = "nodename.local"
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex._container_specjson = dict()
        rcex._container_specjson["root"] = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["linux"] = dict()
        rcex._container_specjson["linux"]["uidMappings"] = dict()
        rcex._container_specjson["linux"]["gidMappings"] = dict()
        rcex.opt["cwd"] = "/"
        rcex.opt["env"] = []
        rcex.opt["cmd"] = "bash"
        rcex.opt["hostname"] = ""
        json_obj = rcex._set_spec()
        self.assertEqual(json_obj["hostname"], "nodename.local")

        # environment passes to container json
        mock_realpath.return_value = "/.udocker/containers/aaaaaa/ROOT"
        mock_node.return_value = "nodename.local"
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex._container_specjson = dict()
        rcex._container_specjson["root"] = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["linux"] = dict()
        rcex._container_specjson["linux"]["uidMappings"] = dict()
        rcex._container_specjson["linux"]["gidMappings"] = dict()
        rcex.opt["cwd"] = "/"
        rcex.opt["env"] = [("AA", "aa"), ("BB", "bb")]
        rcex.opt["cmd"] = "bash"
        rcex.opt["hostname"] = ""
        json_obj = rcex._set_spec()
        self.assertIn("AA=aa", json_obj["process"]["env"])

        # environment syntax error
        mock_realpath.return_value = "/.udocker/containers/aaaaaa/ROOT"
        mock_node.return_value = "nodename.local"
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex._container_specjson = dict()
        rcex._container_specjson["root"] = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["linux"] = dict()
        rcex._container_specjson["linux"]["uidMappings"] = dict()
        rcex._container_specjson["linux"]["gidMappings"] = dict()
        rcex.opt["cwd"] = "/"
        rcex.opt["env"] = [("BB", "bb")]
        rcex.opt["cmd"] = "bash"
        rcex.opt["hostname"] = ""
        json_obj = rcex._set_spec()
        self.assertNotIn("AA=aa", json_obj["process"]["env"])

        # uid and gid mappings
        mock_realpath.return_value = "/.udocker/containers/aaaaaa/ROOT"
        mock_node.return_value = "nodename.local"
        mock_getuid.return_value = 10000
        mock_getgid.return_value = 10000
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex._container_specjson = dict()
        rcex._container_specjson["root"] = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["linux"] = dict()
        rcex._container_specjson["linux"]["uidMappings"] = dict()
        rcex._container_specjson["linux"]["gidMappings"] = dict()
        rcex._container_specjson["linux"]["uidMappings"]["XXX"] = 0
        rcex._container_specjson["linux"]["gidMappings"]["XXX"] = 0
        rcex.opt["cwd"] = "/"
        rcex.opt["env"] = [("AA", "aa"), ("BB", "bb")]
        rcex.opt["cmd"] = "bash"
        rcex.opt["hostname"] = ""
        json_obj = rcex._set_spec()
        self.assertFalse(mock_getuid.called)
        self.assertFalse(mock_getgid.called)

    @patch('udocker.engine.runc.Msg')
    def test_06__uid_check(self, mock_msg):
        """Test06 RuncEngine()._uid_check()."""
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex._uid_check()
        self.assertFalse(mock_msg.called)

        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex.opt["user"] = "root"
        rcex._uid_check()
        self.assertFalse(mock_msg.called)

        mock_msg.level = 3
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex.opt["user"] = "user01"
        rcex._uid_check()
        self.assertTrue(mock_msg.called)

    def test_07__add_capabilities_spec(self):
        """Test07 RuncEngine()._add_capabilities_spec()."""
        rcex = RuncEngine(self.local, self.xmode)
        Config.conf['runc_capabilities'] = ""
        self.assertEqual(rcex._add_capabilities_spec(), None)

        rcex = RuncEngine(self.local, self.xmode)
        Config.conf['runc_capabilities'] = ["CAP_KILL",
                                            "CAP_NET_BIND_SERVICE",
                                            "CAP_CHOWN"]
        rcex._container_specjson = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["process"]["capabilities"] = dict()
        rcex._container_specjson["process"]["capabilities"]["ambient"] = dict()
        rcex._add_capabilities_spec()
        res = rcex._container_specjson["process"]["capabilities"]["ambient"]
        self.assertEqual(res, Config.conf['runc_capabilities'])

    @patch('udocker.engine.runc.HostInfo')
    @patch('udocker.engine.runc.os.minor')
    @patch('udocker.engine.runc.os.major')
    @patch('udocker.engine.runc.stat.S_ISCHR')
    @patch('udocker.engine.runc.stat.S_ISBLK')
    @patch('udocker.engine.runc.os.path.exists')
    @patch('udocker.engine.runc.Msg')
    def test_08__add_device_spec(self, mock_msg, mock_exists,
                                 mock_blk, mock_chr, mock_osmaj,
                                 mock_osmin, mock_hi):
        """Test08 RuncEngine()._add_device_spec()."""
        mock_msg.level = 3
        mock_exists.return_value = False
        rcex = RuncEngine(self.local, self.xmode)
        status = rcex._add_device_spec("/dev/zero")
        self.assertTrue(mock_msg.return_value.err.called)
        self.assertFalse(status)

        mock_msg.level = 3
        mock_exists.return_value = True
        mock_blk.return_value = False
        mock_chr.return_value = False
        rcex = RuncEngine(self.local, self.xmode)
        status = rcex._add_device_spec("/dev/zero")
        self.assertTrue(mock_msg.return_value.err.called)
        self.assertFalse(status)

        mock_msg.level = 3
        mock_exists.return_value = True
        mock_blk.return_value = True
        mock_chr.return_value = False
        mock_osmaj.return_value = 0
        mock_osmin.return_value = 6
        mock_hi.uid = 0
        mock_hi.gid = 0
        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["linux"] = dict()
        rcex._container_specjson["linux"]["devices"] = list()
        status = rcex._add_device_spec("/dev/zero")
        self.assertTrue(status)
        self.assertTrue(mock_osmaj.called)
        self.assertTrue(mock_osmin.called)

    @patch('udocker.engine.runc.NvidiaMode')
    @patch.object(RuncEngine, '_add_device_spec')
    def test_09__add_devices(self, mock_adddecspec, mock_nv):
        """Test09 RuncEngine()._add_devices()."""
        mock_adddecspec.side_effect = [True, True, True]
        mock_nv.return_value.get_mode.return_value = True
        mock_nv.return_value.get_devices.return_value = ['/dev/nvidia', ]
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt = dict()
        rcex.opt["devices"] = ["/dev/open-mx", "/dev/myri0"]
        rcex._add_devices()
        self.assertTrue(mock_adddecspec.call_count, 3)
        self.assertTrue(mock_nv.return_value.get_mode.called)
        self.assertTrue(mock_nv.return_value.get_devices.called)

    def test_10__add_mount_spec(self):
        """Test10 RuncEngine()._add_mount_spec()."""
        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        rcex._add_mount_spec("/HOSTDIR", "/CONTDIR")
        mount = rcex._container_specjson["mounts"][0]
        self.assertEqual(mount["destination"], "/CONTDIR")
        self.assertEqual(mount["source"], "/HOSTDIR")
        self.assertIn("ro", mount["options"])

        # rw
        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        rcex._add_mount_spec("/HOSTDIR", "/CONTDIR", True)
        mount = rcex._container_specjson["mounts"][0]
        self.assertEqual(mount["destination"], "/CONTDIR")
        self.assertEqual(mount["source"], "/HOSTDIR")
        self.assertIn("rw", mount["options"])

    def test_11__del_mount_spec(self):
        """Test11 RuncEngine()._del_mount_spec()."""
        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        mount = {"destination": "/CONTDIR",
                 "type": "none",
                 "source": "/HOSTDIR",
                 "options": ["rbind", "nosuid",
                             "noexec", "nodev",
                             "rw", ], }
        rcex._container_specjson["mounts"].append(mount)
        rcex._del_mount_spec("/HOSTDIR", "/CONTDIR")
        # self.assertEqual(len(rcex._container_specjson["mounts"]), 0)

        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        mount = {"destination": "/XXXX",
                 "type": "none",
                 "source": "/HOSTDIR",
                 "options": ["rbind", "nosuid",
                             "noexec", "nodev",
                             "rw", ], }
        rcex._container_specjson["mounts"].append(mount)
        rcex._del_mount_spec("/HOSTDIR", "/CONTDIR")
        self.assertEqual(len(rcex._container_specjson["mounts"]), 1)

        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        mount = {"destination": "/CONTDIR",
                 "type": "none",
                 "source": "XXXX",
                 "options": ["rbind", "nosuid",
                             "noexec", "nodev",
                             "rw", ], }
        rcex._container_specjson["mounts"].append(mount)
        rcex._del_mount_spec("/HOSTDIR", "/CONTDIR")
        self.assertEqual(len(rcex._container_specjson["mounts"]), 1)

    def test_12__sel_mount_spec(self):
        """Test12 RuncEngine()._sel_mount_spec()."""
        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        mount = {"destination": "/CONTDIR",
                 "type": "none",
                 "source": "xxx",
                 "options": ["rbind", "nosuid",
                             "noexec", "nodev",
                             "rw", ], }
        rcex._container_specjson["mounts"].append(mount)
        status = rcex._sel_mount_spec("/HOSTDIR", "/CONTDIR")
        self.assertEqual(status, None)

        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        mount = {"destination": "/CONTDIR",
                 "type": "none",
                 "source": "/HOSTDIR",
                 "options": ["rbind", "nosuid",
                             "noexec", "nodev",
                             "rw", ], }
        rcex._container_specjson["mounts"].append(mount)
        status = rcex._sel_mount_spec("/HOSTDIR", "/CONTDIR")
        self.assertEqual(status, 0)

    @patch.object(RuncEngine, '_sel_mount_spec')
    def test_13__mod_mount_spec(self, mock_selmount):
        """Test13 RuncEngine()._mod_mount_spec()."""
        optnew = {"options": ["a", "su"]}
        mock_selmount.return_value = None
        rcex = RuncEngine(self.local, self.xmode)
        status = rcex._mod_mount_spec("/HOSTDIR", "/CONTDIR", optnew)
        self.assertFalse(status)

        optnew = {"options": ["a", "su"]}
        mock_selmount.return_value = 0
        rcex = RuncEngine(self.local, self.xmode)
        rcex._container_specjson = dict()
        rcex._container_specjson["mounts"] = []
        mount = {"destination": "/CONTDIR",
                 "type": "none",
                 "source": "/HOSTDIR",
                 "options": ["a"], }
        rcex._container_specjson["mounts"].append(mount)
        status = rcex._mod_mount_spec("/HOSTDIR", "/CONTDIR", optnew)
        self.assertTrue(status)
        self.assertEqual(len(rcex._container_specjson["mounts"]), 1)

    @patch('udocker.engine.runc.Msg')
    @patch('udocker.engine.runc.os.path.isfile')
    @patch('udocker.engine.runc.os.path.isdir')
    @patch.object(RuncEngine, '_add_mount_spec')
    @patch('udocker.engine.runc.FileBind')
    def test_14__add_volume_bindings(self, mock_fbind,
                                     mock_add_mount_spec,
                                     mock_isdir, mock_isfile, mock_msg):
        """Test14 RuncEngine()._add_volume_bindings()."""
        mock_msg.level = 0
        mock_add_mount_spec.side_effect = [None, None]
        rcex = RuncEngine(self.local, self.xmode)
        rcex._filebind = mock_fbind
        rcex._filebind.start.return_value = ("/HOSTDIR", "/CONTDIR")
        rcex.opt["vol"] = list()
        rcex._add_volume_bindings()
        self.assertTrue(mock_add_mount_spec.called)
        self.assertFalse(mock_isdir.called)

        mock_add_mount_spec.side_effect = [None, None]
        mock_isdir.return_value = False
        mock_isfile.return_value = False
        rcex = RuncEngine(self.local, self.xmode)
        rcex._filebind = mock_fbind
        rcex._filebind.start.return_value = ("/HOSTDIR", "/CONTDIR")
        rcex.opt["vol"] = list()
        rcex.opt["vol"].append("/HOSTDIR:/CONTDIR")
        rcex._add_volume_bindings()
        self.assertTrue(mock_isdir.called)
        self.assertTrue(mock_isfile.called)

        mock_add_mount_spec.side_effect = [None, None]
        mock_isdir.return_value = True
        mock_isfile.return_value = False
        rcex = RuncEngine(self.local, self.xmode)
        rcex._filebind = mock_fbind
        rcex._filebind.start.return_value = ("/HOSTDIR", "/CONTDIR")
        rcex.opt["vol"] = list()
        rcex.opt["vol"].append("/HOSTDIR:/CONTDIR")
        rcex._add_volume_bindings()
        self.assertTrue(mock_isdir.called)
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_add_mount_spec.call_count, 2)

        mock_add_mount_spec.side_effect = [None, None]
        mock_isdir.return_value = False
        mock_isfile.return_value = True
        rcex = RuncEngine(self.local, self.xmode)
        Config.conf['sysdirs_list'] = ["/HOSTDIR"]
        rcex._filebind = mock_fbind
        rcex._filebind.start.return_value = ("/HOSTDIR", "/CONTDIR")
        rcex._filebind.set_file.return_value = None
        rcex._filebind.add_file.return_value = None
        rcex.opt["vol"] = list()
        rcex.opt["vol"].append("/HOSTDIR:/CONTDIR")
        rcex._add_volume_bindings()
        self.assertTrue(mock_isdir.called)
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_add_mount_spec.call_count, 1)
        self.assertTrue(rcex._filebind.set_file.called)
        self.assertTrue(rcex._filebind.add_file.called)

    @patch('udocker.engine.runc.Msg')
    def test_15__run_invalid_options(self, mock_msg):
        """Test15 RuncEngine()._run_invalid_options()."""
        mock_msg.level = 0
        rcex = RuncEngine(self.local, self.xmode)
        rcex.opt['netcoop'] = False
        rcex.opt['portsmap'] = True
        rcex._run_invalid_options()
        self.assertTrue(mock_msg.called)

    @patch.object(RuncEngine, '_create_mountpoint')
    @patch('udocker.engine.runc.stat')
    @patch('udocker.engine.runc.FileUtil.chmod')
    @patch('udocker.engine.runc.FileBind')
    @patch('udocker.engine.runc.os.path.basename')
    @patch('udocker.engine.runc.PRootEngine')
    def test_16__proot_overlay(self, mock_proot, mock_base, mock_fbind,
                               mock_chmod, mock_stat, mock_crmpoint):
        """Test16 RuncEngine()._proot_overlay()."""
        self.xmode.get_mode.return_value = "R1"
        rcex = RuncEngine(self.local, self.xmode)
        status = rcex._proot_overlay("P2")
        self.assertFalse(status)

        self.xmode.get_mode.return_value = "R2"
        mock_proot.return_value.select_proot.return_value = None
        mock_base.return_value = "ls"
        mock_chmod.return_value = None
        mock_crmpoint.return_value = True
        mock_stat.return_value.S_IRUSR = 256
        mock_stat.return_value.S_IWUSR = 128
        mock_stat.return_value.S_IXUSR = 64
        rcex = RuncEngine(self.local, self.xmode)
        rcex._filebind = mock_fbind
        rcex._filebind.start.return_value = ("/HOSTDIR", "/CONTDIR")
        rcex._filebind.set_file.return_value = None
        rcex._filebind.add_file.return_value = None
        rcex._container_specjson = dict()
        rcex._container_specjson["process"] = dict()
        rcex._container_specjson["process"]["env"] = []
        rcex._container_specjson["process"]["args"] = []
        status = rcex._proot_overlay("P2")
        self.assertTrue(status)
        self.assertTrue(mock_base.called)
        self.assertTrue(mock_chmod.called)
        self.assertTrue(mock_crmpoint.called)
        self.assertTrue(rcex._filebind.set_file.called)
        self.assertTrue(rcex._filebind.add_file.called)

    @patch('udocker.engine.runc.FileBind')
    @patch('udocker.engine.runc.Unique')
    @patch.object(RuncEngine, '_set_id_mappings')
    @patch.object(RuncEngine, '_del_namespace_spec')
    @patch.object(RuncEngine, '_del_mount_spec')
    @patch.object(RuncEngine, 'run_nopty')
    @patch.object(RuncEngine, 'run_pty')
    @patch.object(RuncEngine, '_run_invalid_options')
    @patch.object(RuncEngine, '_del_mount_spec')
    @patch.object(RuncEngine, '_run_banner')
    @patch.object(RuncEngine, '_add_devices')
    @patch.object(RuncEngine, '_add_volume_bindings')
    @patch.object(RuncEngine, '_set_spec')
    @patch.object(RuncEngine, '_save_spec')
    @patch.object(RuncEngine, '_proot_overlay')
    @patch.object(RuncEngine, '_run_env_set')
    @patch.object(RuncEngine, '_mod_mount_spec')
    @patch.object(RuncEngine, '_add_capabilities_spec')
    @patch.object(RuncEngine, '_uid_check')
    @patch.object(RuncEngine, '_run_env_cleanup_list')
    @patch.object(RuncEngine, '_load_spec')
    @patch.object(RuncEngine, 'select_runc')
    @patch.object(RuncEngine, '_run_init')
    def test_17_run(self, mock_run_init, mock_sel_runc,
                    mock_load_spec, mock_run_env_cleanup_list,
                    mock_uid_check, mock_addspecs, mock_modms,
                    mock_env_set, mock_prooto, mock_savespec,
                    mock_set_spec, mock_add_bindings, mock_add_dev,
                    mock_run_banner, mock_del_mount_spec, mock_inv_opt,
                    mock_pty, mock_nopty, mock_delmnt, mock_delns,
                    mock_setid,
                    mock_unique, mock_fbind):
        """Test17 RuncEngine().run()."""
        mock_run_init.return_value = False
        mock_delmnt.return_value = None
        mock_delns.return_value = None
        mock_setid.return_value = None
        mock_env_set.return_value = None
        rcex = RuncEngine(self.local, self.xmode)
        status = rcex.run("CONTAINERID")
        self.assertEqual(status, 2)

        mock_run_init.return_value = True
        mock_inv_opt.return_value = None
        mock_load_spec.return_value = False
        mock_sel_runc.return_value = None
        mock_load_spec.return_value = False
        mock_delmnt.return_value = None
        mock_delns.return_value = None
        mock_setid.return_value = None
        rcex = RuncEngine(self.local, self.xmode)
        rcex.container_dir = "/container/ROOT"
        rcex._filebind = mock_fbind
        rcex._filebind.setup.return_value = None
        status = rcex.run("CONTAINERID")
        self.assertEqual(status, 4)

        jload = {"container": "cxxx", "parent": "dyyy",
                 "created": "2020-05-05T21:20:07.182447994Z",
                 "os": "linux",
                 "container_config": {"Tty": "false", "Cmd": ["/bin/sh"]},
                 "Image": "sha256:aa"
                }
        mock_run_init.return_value = True
        mock_inv_opt.return_value = None
        mock_load_spec.return_value = False
        mock_sel_runc.return_value = None
        mock_load_spec.return_value = jload
        mock_uid_check.return_value = None
        mock_run_env_cleanup_list.return_value = None
        mock_set_spec.return_value = None
        mock_del_mount_spec.return_value = None
        mock_add_bindings.return_value = None
        mock_add_dev.return_value = None
        mock_addspecs.return_value = None
        mock_modms.return_value = None
        mock_prooto.return_value = None
        mock_savespec.return_value = None
        mock_unique.return_value.uuid.return_value = "EXECUTION_ID"
        mock_run_banner.return_value = None
        mock_pty.return_value = 0
        mock_nopty.return_value = 0
        mock_delmnt.return_value = None
        mock_delns.return_value = None
        mock_setid.return_value = None
        rcex = RuncEngine(self.local, self.xmode)
        rcex.container_dir = "/container/ROOT"
        rcex._filebind = mock_fbind
        rcex._filebind.setup.return_value = None
        status = rcex.run("CONTAINERID")
        self.assertEqual(status, 0)

    @patch('udocker.engine.runc.FileBind')
    @patch('udocker.engine.runc.subprocess.call')
    def test_18_run_pty(self, mock_call, mock_fbind):
        """Test18 RuncEngine().run_pty()."""
        mock_call.return_value = 0
        rcex = RuncEngine(self.local, self.xmode)
        rcex.container_dir = "/container/ROOT"
        rcex._filebind = mock_fbind
        rcex._filebind.finish.return_value = None
        status = rcex.run_pty("CONTAINERID")
        self.assertEqual(status, 0)

    # def test_19_run_nopty(self):
    #     """Test19 RuncEngine().run_nopty()."""
    #     pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""File bind inside containers"""

import os

from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil


class FileBind(object):
    """Alternative method to allow host files to be visible inside
    a container when binding of files is not possible such as when
    using rootless namespaces.
    """

    bind_dir = "/.bind_host_files"
    orig_dir = "/.bind_orig_files"

    def __init__(self, localrepo, container_id):
        self.localrepo = localrepo               # LocalRepository object
        self.container_id = container_id         # Container id
        self.container_dir = \
            os.path.realpath(self.localrepo.cd_container(container_id))
        self.container_root = self.container_dir + "/ROOT"
        self.container_bind_dir = self.container_root + self.bind_dir
        self.container_orig_dir = self.container_dir + self.orig_dir
        self.host_bind_dir = None

    def setup(self):
        """Prepare container for FileBind"""
        if not os.path.isdir(self.container_orig_dir):
            if not FileUtil(self.container_orig_dir).mkdir():
                Msg().err("Error: creating dir:", self.container_orig_dir)
                return False
        if not os.path.isdir(self.container_bind_dir):
            if not FileUtil(self.container_bind_dir).mkdir():
                Msg().err("Error: creating dir:", self.container_bind_dir)
                return False
        return True

    def restore(self, force=False):
        """Restore container files after FileBind"""
        error = False
        if not os.path.isdir(self.container_orig_dir):
            return
        for f_name in os.listdir(self.container_orig_dir):
            orig_file = self.container_orig_dir + '/' + f_name
            if not os.path.isfile(orig_file):
                continue
            cont_file = os.path.basename(f_name).replace('#', '/')
            cont_file = self.container_root + '/' + cont_file
            if os.path.islink(cont_file):
                FileUtil(cont_file).register_prefix()
                FileUtil(cont_file).remove()
            elif os.path.exists(cont_file):
                continue
            if not FileUtil(orig_file).rename(cont_file):
                Msg().err("Error: restoring binded file:", cont_file)
                error = True
        if not error or force:
            FileUtil(self.container_orig_dir).remove(recursive=True)
        FileUtil(self.container_bind_dir).remove(recursive=True)

    def start(self, files_list=None):
        """Prepare host files to be made available inside container
        files_list: is the initial list of files to be made available
        returns: the directory that holds the files and that must be
                 binded inside the container
        """
        self.setup()
        self.host_bind_dir = FileUtil("BIND_FILES").mktmpdir()
        if files_list:
            self.set_list(files_list)
        return (self.host_bind_dir, self.bind_dir)

    def set_list(self, files_list):
        """setup links for a list of files"""
        for f_name in files_list:
            self.set_file(f_name, f_name)

    def set_file(self, host_file, cont_file):
        """Prepare individual file mapping"""
        if not os.path.isfile(host_file):
            return
        orig_file = cont_file.replace('/', '#')
        orig_file_path = self.container_orig_dir + '/' + orig_file
        cont_file_path = self.container_root + '/' + cont_file
        link_path = self.bind_dir + '/' + orig_file
        if not os.path.exists(orig_file_path):
            if os.path.isfile(cont_file_path):
                os.rename(cont_file_path, orig_file_path)
            else:
                return
            os.symlink(link_path, cont_file_path)
        FileUtil(orig_file_path).copyto(self.host_bind_dir)

    def add_file(self, host_file, cont_file):
        """Add file to be made available inside container"""
        replace_file = cont_file.replace('/', '#')
        replace_file = self.host_bind_dir + '/' + replace_file
        FileUtil(replace_file).remove()
        FileUtil(host_file).copyto(replace_file)

    def get_path(self, cont_file):
        """Get real path of file in container self.host_bind_dir"""
        replace_file = cont_file.replace('/', '#')
        return self.host_bind_dir + '/' + replace_file

    def finish(self):
        """Cleanup after run"""
        return
